# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from pyecharts.charts import Bar
from pyecharts.options import (
    AxisOpts,
    InitOpts,
    LabelOpts,
    MarkPointItem,
    MarkPointOpts,
)

from dlk.display import Display, DisplayConfigBase
from dlk.utils.register import register

logger = logging.getLogger(__name__)


@cregister("display", "txt_cls")
class TextClassificationDisplayConfig(DisplayConfigBase):
    """the text classification display"""

    title = StrField("Text Classification Task Demo", help="the title of the demo")
    help = StrField(
        "You can add the sentence to the `sentence input area` and click `submit` to get the class info of this sentence",
        help="the help infomation of the demo",
    )
    render_height = IntField(value=500, help="the display height")
    render_width = IntField(value=900, help="the display width")
    render_type = StrField(value="html", help="the render type of the display")
    output_head = StrField(value="Text Class", help="the output head")

    class Input:
        sentence = StrField(
            "text",
            options=["text"],
            help="the input values name for classification, the value is the type",
        )

    labels = ListField(
        [], help=f"the display labels, if not set, will display all the labels"
    )
    top_k = IntField(
        10, minimum=0, help="the top k class to display, if 0, will display all"
    )

    input = NestField(value=Input, help="the input config")


@cregister("display", "txt_match")
class TextMatchDisplayConfig(TextClassificationDisplayConfig):
    """the text match display"""

    title = StrField("Text Match Task Demo", help="the title of the demo")
    help = StrField(
        "You can add the sentence(a and b) to the `sentence input area(a and b)` and click `submit` to get the match class info of the pair sentences",
        help="the help infomation of the demo",
    )

    class Input:
        sentence_a = StrField("text", options=["text"], help="the text a for match")
        sentence_b = StrField("text", options=["text"], help="the text b for match")

    input = NestField(value=Input, help="the input config")


class TextClassificationDisplay(Display):
    """text classification display"""

    def __init__(self, config: TextClassificationDisplayConfig):
        super(TextClassificationDisplay, self).__init__(config)
        self.config = config
        self.init_options = InitOpts(
            width=f"{config.render_width}px",
            height=f"{config.render_height}px",
        )

    def display(self, result):
        predicts = result["predicts"]
        label_values = {}
        for _, pair in predicts.items():
            label_values[pair[0]] = float(f"{pair[1]*100:.2f}")
        if self.config.top_k > 0:
            label_values = dict(
                sorted(label_values.items(), key=lambda x: x[1], reverse=True)[
                    : self.config.top_k
                ]
            )
        labels = []
        if self.config.labels:
            labels = self.config.labels
        else:
            labels = list(label_values.keys())
            labels.sort()
        return (
            Bar(init_opts=self.init_options)
            .set_global_opts(
                yaxis_opts=AxisOpts(
                    axislabel_opts=LabelOpts(formatter="{value} %"),
                    name="Probability (%)",
                ),
                xaxis_opts=AxisOpts(name="Class", axislabel_opts=LabelOpts(rotate=45)),
            )
            .add_xaxis(labels)
            .add_yaxis(
                "Probability",
                [label_values[label] for label in labels],
                category_gap="50%",
            )
            .set_series_opts(
                label_opts=LabelOpts(is_show=False),
                markpoint_opts=MarkPointOpts(
                    data=[
                        MarkPointItem(type_="max", name="Top 1"),
                    ]
                ),
            )
            .render_embed()
        )


register("display", "txt_cls")(TextClassificationDisplay)
register("display", "txt_match")(TextClassificationDisplay)



# Copyright the author(s) of DLK.

# There are many code copy from fairseq.
# Copyright (c) Facebook, Inc. and its affiliates.

"""sample method"""
import os
from typing import List, Optional

import torch
import torch.nn as nn
from torch import Tensor

from dlk.utils.import_module import import_module_dir
from dlk.utils.token_generation_constraints import (
    ConstraintState,
    OrderedConstraintState,
    UnorderedConstraintState,
)


class Sample(nn.Module):
    def __init__(self, tgt_dict, config=None):
        super().__init__()
        self.pad = tgt_dict.pad()
        self.unk = tgt_dict.unk()
        self.eos = tgt_dict.eos()
        self.vocab_size = len(tgt_dict)
        self.src_lengths = torch.tensor(-1)
        self.supports_constraints = False
        self.tgt_dict = tgt_dict
        self.config = config

    def step(
        self, step, lprobs, scores, prev_output_tokens=None, original_batch_idxs=None
    ):
        """Take a single Sample step.

        Args:
            step: the current Sample step, starting at 0
            lprobs: (bsz x input_beam_size x vocab_size)
                the model's log-probabilities over the vocabulary at the current step
            scores: (bsz x input_beam_size x step)
                the historical model scores of each hypothesis up to this point
            prev_output_tokens: (bsz x step)
                the previously generated oputput tokens
            original_batch_idxs: (bsz)
                the tensor with the batch indices, in the range [0, bsz)
                this is useful in case there has been applied a re-ordering
                and we need to know the original indices

        Return: A tuple of (scores, indices, beams) where:
            scores: (bsz x output_beam_size)
                the scores of the chosen elements; output_beam_size can be
                larger than input_beam_size, e.g., we may return
                2*input_beam_size to account for EOS
            indices: (bsz x output_beam_size)
                the indices of the chosen elements
            beams: (bsz x output_beam_size)
                the hypothesis ids of the chosen elements, in the range [0, input_beam_size)
        """
        raise NotImplementedError

    def set_src_lengths(self, src_lengths):
        self.src_lengths = src_lengths

    def init_constraints(self, batch_constraints: Optional[Tensor], beam_size: int):
        """Initialize constraint states for constrained decoding (if supported).

        Args:
            batch_constraints: (torch.Tensor, optional)
                the list of constraints, in packed form
            beam_size: (int)
                the beam size
        Returns:
            *encoder_out* rearranged according to *new_order*
        """
        pass

    def prune_sentences(self, batch_idxs: Tensor):
        """
        Removes constraint states for completed sentences (if supported).
        This is called from sequence_generator._generate() when sentences are
        deleted from the batch.

        Args:
            batch_idxs: Indices of *sentences* whose constraint state should be *kept*.
        """
        pass

    def update_constraints(self, active_hypos: Tensor):
        """
        Updates the constraint states by selecting the beam items that are retained.
        This is called at each time step of sequence_generator._generate() when
        the set of 2 * {beam_size} candidate hypotheses are reduced to the beam size.

        Args:
            active_hypos: (batch size, beam size)
              list of integers denoting, for each sentence, which beam candidate items
              should be kept.
        """
        pass


sample_dir = os.path.dirname(__file__)
import_module_dir(sample_dir, "dlk.nn.token_sample")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
from typing import Callable, Dict, Iterable, List, Set, Union

import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "char_gather")
class CharGatherConfig(BaseSubProcessorConfig):
    """the char gather subprocessor"""

    train_data_set = ListField(
        value=["train"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )
    gather_columns = ListField(
        value=MISSING,
        help="List of columns. Every cell must be sigle token or list of tokens or set of tokens",
    )
    char_vocab = StrField(
        value="char_vocab.json",
        help="save Vocabulary object (the Vocabulary of labels) file.",
    )
    ignore = StrField(
        value="",
        help="ignore the token, the id of this token will be -100, null means nothing should be ignore",
    )
    update = StrField(
        value="",
        help="null or another exists Vocabulary object should be update, if the update is not null, the unk and pad ignore will be ignored",
    )
    unk = StrField(value="[UNK]", help="the unk token")
    pad = StrField(value="[PAD]", help="the pad token")
    min_freq = IntField(
        value=1,
        minimum=1,
        help="the min freq of token, you can only change one of the value of min_freq and most_common",
    )
    most_common = IntField(
        value=-1,
        minimum=-1,
        help="the most common token, -1 for all, you can only change one of the value of min_freq and most_common.",
    )


@register("subprocessor", "char_gather")
class CharGather(BaseSubProcessor):
    """gather all character from the 'gather_columns' and deliver a vocab named 'char_vocab'"""

    def __init__(self, stage: str, config: CharGatherConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.config = config
        self.update = (
            None
            if not self.config.update
            else os.path.join(self.meta_dir, self.config.update)
        )

    def split_to_char(self, input: Union[str, Iterable]):
        """the char is from token or sentence, so we need split them to List[char]

        Args:
            input: auto detach the type of input and split it to char

        Returns:
            the same shape of the input but the str is split to List[char]

        """
        if isinstance(input, str):
            return [c for c in input]
        else:
            return [self.split_to_char(sub_input) for sub_input in input]

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """Character gather entry

        Args:
            data:
            >>> |sentence |label|
            >>> |---------|-----|
            >>> |sent_a...|la   |
            >>> |sent_b...|lb   |

            deliver_meta:
                if there are some meta info need to deliver to next processor, and deliver_meta is True, save the meta info to datadir
        Returns:
            processed data

        """
        if not deliver_meta:
            return data
        if self.update:
            with open(self.update, mode="r", encoding="utf-8") as f:
                self.vocab = Vocabulary.load(json.load(f))
        else:
            self.vocab = Vocabulary(
                do_strip=True, unknown=self.config.unk, ignore=self.config.ignore
            )
        for column in self.config.gather_columns:
            column: str
            self.vocab.auto_update(self.split_to_char(data[column]))
        self.vocab.filter_rare(self.config.min_freq, self.config.most_common)
        logger.info(f"The Char Vocab Num is {self.vocab.word_num}")
        with open(
            os.path.join(self.meta_dir, self.config.char_vocab), "w", encoding="utf-8"
        ) as f:
            json.dump(self.vocab.dumps(), f)
        return data



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import BaseLoss, BaseLossConfig

logger = logging.getLogger(__name__)


@cregister("loss", "focal_loss")
class FocalLossConfig(BaseLossConfig):
    """the focal_loss loss config"""

    weight = ListField(
        value=None,
        additions=[None],
        help="the list of weights of every class",
    )
    ignore_index = IntField(value=-100, help="the ignore index")
    gamma = FloatField(value=1.0, help="the gamma of focal loss")


@register("loss", "focal_loss")
class FocalLoss(BaseLoss):
    """for multi class classification"""

    def __init__(self, config: FocalLossConfig):
        super(FocalLoss, self).__init__(config)
        self.config: FocalLossConfig
        if self.config.weight:
            self.weight = torch.tensor(self.config.weight, dtype=torch.float)
        else:
            self.weight = None

    def focal_loss(self, pred, target):
        """calc focal loss

        Args:
            pred: NxC for logits
            target: N for target

        Returns:
            loss

        """
        # drop the ignore target
        mask = target != self.config.ignore_index
        target = target[mask]
        pred = pred[mask]
        # [N, 1]
        target = target.unsqueeze(-1)
        # [N, C]
        pt = F.softmax(pred, dim=-1)
        logpt = F.log_softmax(pred, dim=-1)
        # [N]
        pt = pt.gather(1, target).squeeze(-1)
        logpt = logpt.gather(1, target).squeeze(-1)

        if self.weight is not None:
            # [N] at[i] = weight[target[i]]
            at = self.weight.gather(0, target.squeeze(-1))
            logpt = logpt * at

        loss = -1 * (1 - pt) ** self.config.gamma * logpt
        if self.config.reduction == "none":
            return loss
        if self.config.reduction == "mean":
            return loss.mean()
        return loss.sum()

    def _calc(self, result, inputs, rt_config, scale):
        """calc the loss the predict is from result, the ground truth is from inputs

        Args:
            result: the model predict dict
            inputs: the all inputs for model
            rt_config: provide the current training status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            scale: the scale rate for the loss

        Returns:
            loss

        """
        pred = result[self.pred_name]
        target = inputs[self.truth_name]
        pred = pred.reshape(-1, pred.shape[-1])
        target = target.reshape(-1)

        loss = self.focal_loss(pred, target) * scale
        return loss, {self.config.log_map.loss: loss}



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import abc
import os

from dlk.utils.import_module import import_module_dir

# automatically import any Python files in the models directory
dataset_dir = os.path.dirname(__file__)
import_module_dir(dataset_dir, "dlk.data.dataset")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import pickle as pkl
from typing import Callable, Dict, List, Set

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import SimpleModule
from dlk.utils.register import register, register_module_name


@cregister("embedding", "combine_word_char_cnn")
class CombineWordCharCNNEmbeddingConfig(Base):
    """the combine_word_char_cnn embedding module config"""

    char_embedding_dim = IntField(value=MISSING, minimum=0, help="the embedding dim")
    word_embedding_dim = IntField(value=MISSING, minimum=0, help="the embedding dim")
    char_embedding_file = StrField(
        value="char_embedding", help="the char embedding file"
    )
    word_embedding_file = StrField(
        value="word_embedding", help="the word embedding file"
    )
    dropout = FloatField(value=0, minimum=0.0, maximum=1.0, help="dropout rate")

    class InputMap:
        char_ids = StrField(value="char_ids", help="the char ids")
        input_ids = StrField(value="input_ids", help="the input ids")

    input_map = NestField(
        value=InputMap, help="the input map of the combine embedding module"
    )

    class OutputMap:
        embedding = StrField(value="embedding", help="the embedding")

    output_map = NestField(
        value=OutputMap, help="the output map of the combine embedding module"
    )

    submodule = SubModule(
        value={
            "embedding#char": {
                "_base": "static_char_cnn",
                "embedding_file": "@lambda @$$.char_embedding_file",
                "dropout": "@lambda @$$.dropout",
                "embedding_dim": "@lambda @$$.char_embedding_dim",
                "output_map": {"char_embedding": "char_embedding"},
            },
            "embedding#word": {
                "_base": "static",
                "embedding_file": "@lambda @$$.word_embedding_file",
                "dropout": "@lambda @$$.dropout",
                "embedding_dim": "@lambda @$$.word_embedding_dim",
                "output_map": {"embedding": "word_embedding"},
            },
        },
        help="the char and word embedding module",
    )


@register("embedding", "combine_word_char_cnn")
class CombineWordCharCNNEmbedding(SimpleModule):
    """from 'input_ids' and 'char_ids' generate 'embedding'"""

    def __init__(self, config: CombineWordCharCNNEmbeddingConfig):
        super().__init__(config)
        self.dropout = nn.Dropout(float(config.dropout))
        word_embedding_config = config["@embedding#word"]
        self.word_embedding = register.get(
            "embedding", register_module_name(word_embedding_config._module_name)
        )(word_embedding_config)
        char_embedding_config = config["@embedding#char"]
        self.char_embedding = register.get(
            "embedding", register_module_name(char_embedding_config._module_name)
        )(char_embedding_config)
        self.config = config

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        self.word_embedding.init_weight(method)
        self.char_embedding.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """get the combine char and word embedding

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        inputs = self.word_embedding(inputs)
        inputs = self.char_embedding(inputs)

        combine_embedding = torch.cat(
            [
                inputs[self.char_embedding.config.output_map.char_embedding],
                inputs[self.word_embedding.config.output_map.embedding],
            ],
            dim=-1,
        )
        inputs[self.config.output_map.embedding] = self.dropout(combine_embedding)
        return inputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict

import torch.nn as nn
import torch.optim as optim
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

from . import BaseOptimizer, BaseOptimizerConfig


@cregister("optimizer", "adamw")
class AdamWOptimizerConfig(BaseOptimizerConfig):
    """the AdamW optimizer"""

    lr = FloatField(value=1e-3, minimum=0.0, help="the learning rate of optimizer")
    betas = ListField(value=[0.9, 0.999], help="the betas of adamw")
    eps = FloatField(value=1e-8, minimum=1e-20, help="the epsilon of optimizer")
    weight_decay = FloatField(
        value=1e-2, minimum=0.0, help="the weight decay of the optimizer"
    )


@register("optimizer", "adamw")
class AdamWOptimizer(BaseOptimizer):
    """Wrap for optim.AdamW"""

    def __init__(self, model: nn.Module, config: AdamWOptimizerConfig):
        super(AdamWOptimizer, self).__init__(model, config, optim.AdamW)



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Dict, List, Set

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import SimpleModule
from dlk.utils.register import register, register_module_name


@cregister("decoder", "multi_decoder")
class MultiDecoderConfig(Base):
    """for multiple decoder"""

    submodule = SubModule({}, help="all the decode module config")


@register("decoder", "multi_decoder")
class MultiDecoder(SimpleModule):
    """multi_decoder a x A x b"""

    def __init__(self, config: MultiDecoderConfig):
        super(MultiDecoder, self).__init__(config)
        decode_configs = config._get_named_modules("decoder")
        self.decoders = nn.ModuleDict(
            {
                decode_name: register.get(
                    "decoder", register_module_name(decode_config._module_name)
                )(decode_config)
                for decode_name, decode_config in decode_configs.items()
            }
        )

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns:
            None

        """
        for decode_name in self.decoders:
            self.decoders[decode_name].init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """

        Args:
            inputs: one mini-batch inputs

        Returns:
            one mini-batch outputs

        """
        for decode_name in self.decoders:
            inputs = self.decoders[decode_name](inputs)
        return inputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register, register_module_name


@cregister("loss", "multi_loss")
class MultiLossConfig(Base):
    """multiple loss"""

    loss_collect = StrField(
        value="sum",
        help="the method to collect losses",
    )

    class LogMap:
        loss = StrField(value="loss", help="the output loss name")

    log_map = NestField(value=LogMap, help="the output loss name")
    submodule = SubModule(value={}, help="the multible loss modules")


@register("additional_loss_collect", "sum")
def loss_sum(losses: Dict[str, torch.Tensor], **args: Dict):
    """sum all losses
    Args:
        losses (List): list of loss
    Returns:
        sum of losses
    """
    loss = sum([losses[key] for key in losses])
    return loss


@register("loss", "multi_loss")
class MultiLoss(nn.Module):
    """This module is NotImplemented yet don't use it"""

    def __init__(self, config: MultiLossConfig):
        super(MultiLoss, self).__init__()
        self.config = config
        self.loss_collect = register.get(
            "additional_loss_collect", self.config.loss_collect
        )
        loss_configs = config._get_named_modules("loss")

        self.losses = nn.ModuleDict(
            {
                loss_name: register.get(
                    "loss", register_module_name(loss_config._module_name)
                )(loss_config)
                for loss_name, loss_config in loss_configs.items()
            }
        )

    def update_config(self, rt_config):
        """callback for imodel to update the total steps and epochs

        when init the loss module, the total step and epoch is not known, when all data ready, the imodel update the value for loss module

        Args:
            rt_config: { "total_steps": self.num_training_steps, "total_epochs": self.num_training_epochs}

        Returns:
            None

        """
        for _, loss_module in self.losses.items():
            loss_module.update_config(rt_config)

    def _calc(self, result, inputs, rt_config):
        """calc the loss the predict is from result, the ground truth is from inputs

        Args:
            result: the model predict dict
            inputs: the all inputs for model
            rt_config: provide the current training status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            loss

        """
        losses = {}
        log_loss = {}
        for loss_module in self.losses:
            loss, log = self.losses[loss_module](result, inputs, rt_config)
            losses[loss_module] = loss
            log_loss.update(log)
        loss = self.loss_collect(losses=losses, rt_config=rt_config)
        log_loss.update({self.config.log_map.loss: loss})
        return loss, log_loss

    def __call__(self, result, inputs, rt_config):
        """same as self.calc"""
        return self._calc(result, inputs, rt_config)



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from functools import partial
from typing import Callable, Dict, List, Set

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "token2id")
class Token2IDConfig(BaseSubProcessorConfig):
    """the token 2 character id subprocessor"""

    train_data_set = ListField(
        value=["train", "valid", "test"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )
    predict = ListField(
        value=["predict"],
        suggestions=[["predict"]],
        help="the data set should be processed for predict stage",
    )
    online = ListField(
        value=["online"],
        suggestions=[["online"]],
        help="the data set should be processed for online stage",
    )

    class InputMap:
        tokens = StrField(
            value="tokens",
            suggestions=["tokens", "labels"],
            help="the tokens",
        )

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OutputMap:
        token_ids = StrField(
            value="token_ids",
            suggestions=["token_ids", "label_ids"],
            help="the token ids",
        )

    output_map = NestField(
        value=OutputMap,
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )
    vocab = StrField(
        value="token_vocab.json",
        suggestions=["token_vocab.json", "label_vocab.json"],
        help="the vocab for the token",
    )


@cregister("subprocessor", "token2id-label")
class Token2IDLabelConfig(Token2IDConfig):
    """the token 2 character id subprocessor, generally used for label to id"""

    predict = ListField(
        value=[],
        suggestions=[["predict"]],
        help="the data set should be processed for predict stage",
    )
    online = ListField(
        value=[],
        suggestions=[["online"]],
        help="the data set should be processed for online stage",
    )

    class InputMap:
        tokens = StrField(
            value="labels",
            suggestions=["tokens", "labels"],
            help="the tokens",
        )

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OutputMap:
        token_ids = StrField(
            value="label_ids",
            suggestions=["token_ids", "label_ids"],
            help="the token ids",
        )

    output_map = NestField(
        value=OutputMap,
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )
    vocab = StrField(
        value="label_vocab.json",
        suggestions=["token_vocab.json", "label_vocab.json"],
        help="the vocab for the token",
    )


@register("subprocessor", "token2id")
class Token2ID(BaseSubProcessor):
    """Use 'Vocabulary' map the tokens to id"""

    def __init__(self, stage: str, config: Token2IDConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.stage = stage
        self.config = config
        self.vocab: Vocabulary = None

    def load_meta(self):
        self.loaded_meta = True
        self.vocab = Vocabulary.load_from_file(
            os.path.join(self.meta_dir, self.config.vocab)
        )

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """firstpiece relabel the data

        Args:
            data: will processed data

            deliver_meta:
                ignore
        Returns:
            relabeld data
        """
        if not self.loaded_meta:
            self.load_meta()

        def get_index_wrap(key, x):
            return self.vocab.auto_get_index(x[key])

        get_index = partial(get_index_wrap, self.config.input_map.tokens)
        data[self.config.output_map.token_ids] = data.apply(get_index, axis=1)
        return data



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
from functools import partial

import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from tokenizers import Tokenizer

from dlk.utils.io import open
from dlk.utils.register import register

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "token_norm")
class TokenNormConfig(BaseSubProcessorConfig):
    """the token norm subprocessor"""

    train_data_set = ListField(
        value=["train", "valid", "test"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )
    predict_data_set = ListField(
        value=["predict"],
        suggestions=[["predict"]],
        help="the data set should be processed for predict stage",
    )
    online_data_set = ListField(
        value=["online"],
        suggestions=[["online"]],
        help="the data set should be processed for online stage",
    )

    zero_digits_replaced = BoolField(
        value=True, help="replace the digits to 0, like 1234 -> 0000"
    )
    lowercase = BoolField(value=True, help="lowercase the tokens")
    tokenizer_path = StrField(
        value="tokenizer.json",
        suggestions=["tokenizer.json"],
        help="the tokenizer of tokens",
    )

    class InputMap:
        sentence = StrField(
            value="sentence",
            suggestions=["sentence"],
            help="the input sentence",
        )

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OutputMap:
        norm_sentence = StrField(
            value="norm_sentence",
            suggestions=["norm_sentence"],
            help="the normed sentence",
        )

    output_map = NestField(
        value=OutputMap,
        help="the output map of the processor, the key is the name of the processor provided key, the value is the nexted processor needed key",
    )


@register("subprocessor", "token_norm")
class TokenNorm(BaseSubProcessor):
    """
    This part could merged to fast_tokenizer(it will save some time), but not all process need this part(except some special dataset like conll2003), and will make the fast_tokenizer be heavy.

    Token norm:
        Love -> love
        3281 -> 0000
    """

    def __init__(self, stage: str, config: TokenNormConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.stage = stage
        self.config = config

        with open(os.path.join(self.config.tokenizer_path), "r", encoding="utf-8") as f:
            tokenizer_str = json.dumps(json.load(f))

        self.tokenizer = Tokenizer.from_str(tokenizer_str)
        self.vocab = self.tokenizer.get_vocab()
        self.prefix = self.tokenizer.model.continuing_subword_prefix
        self.unk = self.tokenizer.model.unk_token

        self._zero_digits_replaced_num = 0
        self._lower_case_num = 0
        self._lower_case_zero_digits_replaced_num = 0
        logger.info(
            f"We use zero digits to replace digit token num is {self._zero_digits_replaced_num}, do lowercase token num is {self._lower_case_num}, do both num is {self._lower_case_zero_digits_replaced_num}"
        )

    def token_norm(self, token: str) -> str:
        """norm token, the result len(result) == len(token), exp.  12348->00000

        Args:
            token: origin token

        Returns:
            normed_token

        """
        if token in self.vocab:
            return token

        if self.config.zero_digits_replaced:
            norm = ""
            digit_num = 0
            for c in token:
                if c.isdigit() or c == ".":
                    norm += "0"
                    digit_num += 1
                else:
                    norm += c
            if norm in self.vocab or self.prefix + norm in self.vocab:
                self._zero_digits_replaced_num += 1
                return norm

        if self.config.lowercase:
            norm = token.lower()
            if norm in self.vocab or self.prefix + norm in self.vocab:
                self._lower_case_num += 1
                return norm

        if self.config.lowercase and self.config.zero_digits_replaced:
            norm = ""
            for c in token.lower():
                if c.isdigit() or c == ".":
                    norm += "0"
                else:
                    norm += c
            if norm in self.vocab or self.prefix + norm in self.vocab:
                self._lower_case_zero_digits_replaced_num += 1
                return norm
        return ""

    def seq_norm(self, key: str, one_item: pd.Series) -> str:
        """norm a sentence, the sentence is from one_item[key]

        Args:
            key: the name in one_item
            one_item: a pd.Series which include the key

        Returns:
            norm_sentence

        """
        seq = one_item[key]
        norm_seq = [c for c in seq]
        encode = self.tokenizer.encode(seq)
        for i, token in enumerate(encode.tokens):
            if token == self.unk:
                token_offset = encode.offsets[i]
                prenorm_token: str = seq[token_offset[0] : token_offset[1]]
                norm_token = self.token_norm(prenorm_token)
                if not norm_token:
                    continue
                assert (
                    len(norm_token) == token_offset[1] - token_offset[0]
                ), f"Prenorm '{prenorm_token}', postnorm: '{norm_token}' and {len(norm_token)}!= {token_offset[1]} - {token_offset[0]}"
                norm_seq[token_offset[0] : token_offset[1]] = norm_token
        return "".join(norm_seq)

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """Character gather entry

        Args:
            data:
            >>> |sentence |label|
            >>> |---------|-----|
            >>> |sent_a...|la   |
            >>> |sent_b...|lb   |

            deliver_meta:
                if there are some meta info need to deliver to next processor, and deliver_meta is True, save the meta info to datadir
        Returns:
            processed data

        """
        _seq_norm = partial(self.seq_norm, self.config.input_map.sentence)
        data[self.config.output_map.norm_sentence] = data.apply(_seq_norm, axis=1)
        # WARNING: if you change the apply to parallel_apply, you should change the _zero_digits_replaced_num, etc. to multiprocess safely(BTW, use parallel_apply in tokenizers==0.10.3 will make the process very slow)
        # data_set[value] = data_set.apply(_seq_norm, axis=1)

        return data



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.data.postprocessor import BasePostProcessor, BasePostProcessorConfig
from dlk.data.postprocessor.txt_cls import (
    TxtClsPostProcessor,
    TxtClsPostProcessorConfig,
)
from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

logger = logging.getLogger(__name__)


@cregister("postprocessor", "img_cls")
class ImgClsPostProcessorConfig(TxtClsPostProcessorConfig):
    """image classfication postprocessor"""

    data_type = StrField(
        value="single", options=["single"], help="the data type, only single support"
    )

    class OriginInputMap:
        uuid = StrField(value="uuid", help="the uuid or the id of the sample")
        image = StrField(value="image", help="the image of the sample")
        image_url = StrField(value="", help="the url of the image, default is empty")

    origin_input_map = NestField(
        value=OriginInputMap,
        help="the origin input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )


@register("postprocessor", "img_cls")
class ImgClsPostProcessor(TxtClsPostProcessor):
    """postprocess for text classfication"""

    def __init__(self, config: ImgClsPostProcessorConfig):
        super(ImgClsPostProcessor, self).__init__(config)
        self.config = config

    def _get_origin_data(self, one_origin: pd.Series) -> Dict:
        """

        Args:
            one_origin: the original data

        Returns:
            the gather origin data

        """
        origin = {}
        if self.config.origin_input_map.image_url:
            origin["image_url"] = one_origin[self.config.origin_input_map.image_url]
        origin["uuid"] = one_origin[self.config.origin_input_map.uuid]
        return origin



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Dict, List

import numpy as np
import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.register import register

logger = logging.getLogger(__name__)


@cregister("initmethod", "default")
class DefaultInitConfig:
    """
    The default init method for the modules
    """

    pass


@register("initmethod", "default")
class DefaultInit(object):
    """default method for init the modules"""

    def __init__(self, config: DefaultInitConfig):
        super().__init__()

    def __call__(self, module):
        """Initialize the weights"""
        if isinstance(module, nn.Linear):
            # use the default kaiming init method
            torch.nn.init.xavier_uniform_(module.weight)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            torch.nn.init.xavier_uniform_(module.weight)
        elif isinstance(module, nn.LayerNorm):
            # use the default init
            # weight are initialized to 1, bias to 0
            module.reset_parameters()
        elif isinstance(module, nn.Conv1d):
            torch.nn.init.kaiming_uniform_(module.weight)
        elif isinstance(module, nn.Conv2d):
            torch.nn.init.kaiming_uniform_(module.weight)
        elif isinstance(module, nn.Conv3d):
            torch.nn.init.kaiming_uniform_(module.weight)
        elif isinstance(module, nn.LSTM):
            self.init_lstm(module)
        elif isinstance(module, nn.ModuleList) or isinstance(module, nn.Sequential):
            for sub_module in module:
                self(sub_module)
        elif isinstance(module, nn.ModuleDict):
            for sub_module in module:
                self(module[sub_module])
        else:
            logger.info(f"{module} is not initialization.")

    def init_lstm(self, lstm):
        """
        Initialize lstm
        """
        for ind in range(0, lstm.num_layers):
            weight = eval("lstm.weight_ih_l" + str(ind))
            bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
            nn.init.uniform_(weight, -bias, bias)
            weight = eval("lstm.weight_hh_l" + str(ind))
            bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
            nn.init.uniform_(weight, -bias, bias)
        if lstm.bidirectional:
            for ind in range(0, lstm.num_layers):
                weight = eval("lstm.weight_ih_l" + str(ind) + "_reverse")
                bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
                nn.init.uniform_(weight, -bias, bias)
                weight = eval("lstm.weight_hh_l" + str(ind) + "_reverse")
                bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
                nn.init.uniform_(weight, -bias, bias)

        if lstm.bias:
            for ind in range(0, lstm.num_layers):
                weight = eval("lstm.bias_ih_l" + str(ind))
                weight.data.zero_()
                weight.data[lstm.hidden_size : 2 * lstm.hidden_size] = 1
                weight = eval("lstm.bias_hh_l" + str(ind))
                weight.data.zero_()
                weight.data[lstm.hidden_size : 2 * lstm.hidden_size] = 1
            if lstm.bidirectional:
                for ind in range(0, lstm.num_layers):
                    weight = eval("lstm.bias_ih_l" + str(ind) + "_reverse")
                    weight.data.zero_()
                    weight.data[lstm.hidden_size : 2 * lstm.hidden_size] = 1
                    weight = eval("lstm.bias_hh_l" + str(ind) + "_reverse")
                    weight.data.zero_()
                    weight.data[lstm.hidden_size : 2 * lstm.hidden_size] = 1



# Copyright cstsunfu.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
from typing import Callable, Dict, Iterable, List, Set, Union

import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

from . import BaseSubProcessor, BaseSubProcessorConfig

logger = logging.getLogger(__name__)


@cregister("subprocessor", "token_gather")
class TokenGatherConfig(BaseSubProcessorConfig):
    """the token gather subprocessor"""

    train_data_set = ListField(
        value=["train"],
        suggestions=[["train", "valid", "test"]],
        help="the data set should be processed for train stage",
    )
    gather_columns = ListField(
        value=MISSING,
        suggestions=[
            ["tokens"],
            [
                "tokens",
                {"column": "entities_column", "trace": "entities_info.labels"},
            ],
        ],
        help="List of columns. If the column is a reprent as a dict,  we will trace the real elements by 'trace'. for example: {'entities_info': [{'start': 1， 'end': 2, labels: ['Label1']}, ..]}, the trace to labels is 'entities_info.labels'",
    )
    token_vocab = StrField(
        value="token_vocab.json",
        help="save Vocabulary object (the Vocabulary of labels) file.",
    )
    ignore = StrField(
        value="",
        help="ignore the token, the id of this token will be -100, null means nothing should be ignore",
    )
    update = StrField(
        value="",
        help="null or another exists Vocabulary object should be update, if the update is not null, the unk and pad ignore will be ignored",
    )
    unk = StrField(value="[UNK]", additions=[None], help="the unk token")
    pad = StrField(value="[PAD]", additions=[None], help="the pad token")
    min_freq = IntField(
        value=1,
        minimum=1,
        help="the min freq of token, you can only change one of the value of min_freq and most_common",
    )
    most_common = IntField(
        value=-1,
        minimum=-1,
        help="the most common token, -1 for all, you can only change one of the value of min_freq and most_common.",
    )


@register("subprocessor", "token_gather")
class TokenGather(BaseSubProcessor):
    """gather all tokens from the 'gather_columns' and deliver a vocab named 'token_vocab'"""

    def __init__(self, stage: str, config: TokenGatherConfig, meta_dir: str):
        super().__init__(stage, config, meta_dir)
        self.stage = stage
        self.config = config
        self.update = (
            None
            if not self.config.update
            else os.path.join(self.meta_dir, self.config.update)
        )

    def get_elements_from_series_by_trace(self, data: pd.Series, trace: str) -> List:
        """get the data from data[trace_path]
        >>> for example:
        >>> data[0] = {'entities_info': [{'start': 0, 'end': 1, 'labels': ['Label1']}]} // data is a series, and every element is as data[0]
        >>> trace = 'entities_info.labels'
        >>> return_result = [['Label1']]

        Args:
            data: origin data series
            trace: get data element trace

        Returns:
            the data in the tail of traces

        """

        def get_elements_from_iter_by_trace(iter: Iterable, cur_trace_list: List):
            if not cur_trace_list:
                return iter
            if isinstance(iter, dict):
                return get_elements_from_iter_by_trace(
                    iter[cur_trace_list[0]], cur_trace_list[1:]
                )
            if isinstance(iter, list) or isinstance(iter, tuple):
                return [
                    get_elements_from_iter_by_trace(sub_iter, cur_trace_list)
                    for sub_iter in iter
                ]
            raise PermissionError(
                f"The trace path is only support type list and dict, but you provide {type(iter)}"
            )

        return [get_elements_from_iter_by_trace(one, trace.split(".")) for one in data]

    def process(self, data: pd.DataFrame, deliver_meta: bool) -> pd.DataFrame:
        """Character gather entry

        Args:
            data:
            >>> |sentence |label|
            >>> |---------|-----|
            >>> |sent_a...|la   |
            >>> |sent_b...|lb   |

            deliver_meta:
                if there are some meta info need to deliver to next processor, and deliver_meta is True, save the meta info to datadir
        Returns:
            processed data

        """
        if not deliver_meta:
            return data
        if self.update:
            with open(self.update, mode="r", encoding="utf-8") as f:
                self.vocab = Vocabulary.load(json.load(f))
        else:
            self.vocab = Vocabulary(
                do_strip=True, unknown=self.config.unk, ignore=self.config.ignore
            )
        for column in self.config.gather_columns:
            if isinstance(column, str):
                self.vocab.auto_update(data[column])
            elif isinstance(column, dict):
                self.vocab.auto_update(
                    self.get_elements_from_series_by_trace(
                        data[column["column"]], trace=column["trace"]
                    )
                )
            else:
                raise PermissionError(
                    f"The gather column currently is only support str or dict."
                )
        self.vocab.filter_rare(self.config.min_freq, self.config.most_common)
        logger.info(f"The Vocab Num is {self.vocab.word_num}")
        with open(
            os.path.join(self.meta_dir, self.config.token_vocab), "w", encoding="utf-8"
        ) as f:
            json.dump(self.vocab.dumps(), f)

        return data



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Callable, Dict, Tuple

from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import BaseIdentityModule
from dlk.utils.import_module import import_module_dir
from dlk.utils.register import register


@cregister("decoder", "identity")
class IdentityModuleConfig:
    """identity decoder"""

    pass


@register("decoder", "identity")
class IdentityModule(BaseIdentityModule):
    pass


decoder_dir = os.path.dirname(__file__)
import_module_dir(decoder_dir, "dlk.nn.layer.decoder")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from io import BytesIO

import fsspec
import pyarrow.parquet as pq

open = fsspec.open


def ls(path: str):
    fs_ins, prefix, path, fs_type = _get_fs_protocol(path)
    if fs_type == "file":
        return fs_ins.ls(path)
    elif fs_type == "hdfs":
        path_infos = fs_ins.ls(path)
        paths = []
        for path_info in path_infos:
            path = prefix + path_info["name"]
            if path_info.get("type") == "directory":
                path += "/"
            paths.append(path)
        return paths
    else:
        raise NotImplementedError(f"Currently, only support hdfs and file protocol.")


def mkdir(path: str, create_parents: bool = True, exist_ok: bool = False):
    fs_ins, prefix, path, fs_type = _get_fs_protocol(path)
    if fs_ins.exists(path) and not exist_ok:
        raise FileExistsError(
            f"{path} already exists. Set exist_ok to True to create it."
        )
    else:
        fs_ins.mkdir(path, create_parents)


def rm(path: str, recursive: bool = False):
    fs_ins, prefix, path, fs_type = _get_fs_protocol(path)
    if path.endswith("/") and not recursive:
        raise PermissionError(
            f"rm: cannot remove '{path}': Is a directory. Set recursive to True to remove it."
        )
    else:
        fs_ins.rm(path, recursive=recursive)


def read_parquet_meta(path: str):
    with open(path, "rb") as f:
        return pq.read_metadata(f)


def read_byte_buffer(path: str):
    with open(path, "rb") as f:
        return BytesIO(f.read())


def _get_fs_protocol(path: str):
    fs_ins, path = fsspec.core.url_to_fs(path)
    fs_ins: fsspec.AbstractFileSystem
    prefix = ""
    fs_type = "file"
    if (
        isinstance(fs_ins.protocol, tuple) and "hdfs" in fs_ins.protocol
    ) or fs_ins.protocol == "hdfs":
        prefix = "hdfs://" + fs_ins.host
        fs_type = "hdfs"
    else:
        assert (
            isinstance(fs_ins.protocol, str) and fs_ins.protocol == "file"
        ), "Currently, only support hdfs and file protocol."
    return fs_ins, prefix, path, fs_type



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Dict

import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    dataclass,
)

from dlk.utils.import_module import import_module_dir


@dataclass
class BaseLossConfig(Base):
    """the base loss config"""

    schedule = ListField(value=[1], help="the schedule of the loss, works with scale")
    scale = ListField(value=[1], help="the scale of the loss for every schedule stage")
    pred_truth_pair = DictField(
        value={},
        suggestions=[{"logits": "label_ids"}],
        help="""
        it's a dict of [predit_logits, truth_target] pair.
        If you have more than one pair,
        you should use the `multi_loss` module and provide this module as the submodule.
        """,
    )
    reduction = StrField(
        value="mean",
        options=["mean", "sum", "none"],
        help="the reduction method, support 'mean', 'sum', 'none'",
    )

    class LogMap:
        loss = StrField(value="loss", help="the output loss name")

    log_map = NestField(value=LogMap, help="the output loss name")


class BaseLoss(nn.Module):
    def __init__(self, config: BaseLossConfig):
        super(BaseLoss, self).__init__()
        self.config = config
        if len(self.config.pred_truth_pair) == 1:
            self.pred_name = list(self.config.pred_truth_pair.keys())[0]
            self.truth_name = self.config.pred_truth_pair[self.pred_name]
        else:
            assert len(self.config.pred_truth_pair) < 1

    def update_config(self, rt_config: Dict):
        """callback for imodel to update the total steps and epochs

        when init the loss module, the total step and epoch is not known, when all data ready, the imodel update the value for loss module

        Args:
            rt_config: { "total_steps": self.num_training_steps, "total_epochs": self.num_training_epochs}

        Returns:
            None

        """
        self.current_stage = 0
        self.config.schedule = [
            rt_config["total_steps"] * i for i in self.config.schedule
        ]

    def _calc(self, result, inputs, rt_config, scale):
        """calc the loss the predict is from result, the ground truth is from inputs

        Args:
            result: the model predict dict
            inputs: the all inputs for model
            rt_config: provide the current training status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            scale: the scale rate for the loss

        Returns:
            loss

        """
        raise NotImplementedError

    def forward(self, result, inputs, rt_config):
        """same as self.calc"""
        if rt_config["current_step"] > self.config.schedule[self.current_stage]:
            self.current_stage += 1
        scale = self.config.scale[self.current_stage]
        return self._calc(result, inputs, rt_config, scale)


loss_dir = os.path.dirname(__file__)
import_module_dir(loss_dir, "dlk.loss")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import abc
import importlib
import json
import logging
import os
from typing import Callable, Dict, List, Type, TypeVar, Union

import pandas as pd
import pyarrow.parquet as pq
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    dataclass,
)

from dlk.utils.import_module import import_module_dir

logger = logging.getLogger(__name__)


@dataclass
class BasePostProcessorConfig(Base):
    """the base postprocessor"""

    meta_dir = StrField(
        value="data/meta_data",
        help="the save dir of the meta info",
    )
    save_root_path = StrField(
        value=None,
        additions=[None],
        help="the root path of the save data, default save to the log dir",
    )
    save_dir = DictField(
        value={
            "valid": "valid_output",
            "test": "test_output",
            "predict": "predict_output",
        },
        help="the save path of the data, relative to the save_root_path",
    )
    start_save_step = IntField(
        value=0, minimum=-1, help="the start save step, -1 means the last step"
    )
    start_save_epoch = IntField(
        value=-1, minimum=-1, help="the start save epoch, -1 means the last epoch"
    )
    predict_extend_return = DictField(value={}, help="the extend return of predict")


class BasePostProcessor(object):
    """the base postprocessor"""

    def __init__(self, config: BasePostProcessorConfig):
        super(BasePostProcessor, self).__init__()
        self.config = config

    def loss_name_map(self, stage) -> str:
        """get the stage loss name

        Args:
            stage: valid, train or test

        Returns:
            loss_name

        """
        map = {
            "valid": "val",
            "train": "train",
            "test": "test",
        }
        return map.get(stage, stage)

    def gather_predict_extend_data(
        self, input_data: Dict, i: int, predict_extend_return: Dict
    ):
        """gather the data register in `predict_extend_return`
        Args:
            input_data:
                the model output
            i:
                the index is i
            predict_extend_return:
                the name map which will be reserved
        Returns:
            a dict of data in input_data which is register in predict_extend_return
        """
        result = {}
        for key, name in predict_extend_return.items():
            data = input_data[name][i]
            if torch.is_tensor(data):
                data = data.detach().tolist()
            result[key] = data
        return result

    def average_loss(self, list_batch_outputs: List[Dict]) -> Dict[str, float]:
        """average all the loss of the list_batches

        Args:
            list_batch_outputs: a list of outputs

        Returns:
            average_loss

        """
        loss_names = []
        average_losses = {}
        batch_num = len(list_batch_outputs)
        if not batch_num:
            return average_losses
        for key in list_batch_outputs[0]:
            if key.endswith("_loss") or key == "loss":
                loss_names.append(key)
                average_losses[key] = 0
        for batch_output in list_batch_outputs:
            for name in loss_names:
                average_losses[name] = average_losses[name] + batch_output.get(name, 0)
        average_losses = {
            key: value / batch_num for key, value in average_losses.items()
        }
        return average_losses

    @abc.abstractmethod
    def do_predict(
        self,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> List:
        """Process the model predict to human readable format

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            all predicts

        """
        raise NotImplementedError

    @abc.abstractmethod
    def do_calc_metrics(
        self,
        predicts: List,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>>
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            the named scores

        """
        raise NotImplementedError

    def do_save(
        self,
        predicts: List,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
        save_condition: bool = False,
    ):
        """save the predict when save_condition==True

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            save_condition: True for save, False for depend on rt_config

        Returns:
            None

        """
        if self.config.start_save_epoch == -1 or self.config.start_save_step == -1:
            self.config.start_save_step = rt_config.get("total_steps", 0) - 1
            self.config.start_save_epoch = rt_config.get("total_epochs", 0) - 1
        if not save_condition and (
            rt_config["current_step"] >= self.config.start_save_step
            or rt_config["current_epoch"] >= self.config.start_save_epoch
        ):
            save_condition = True
        if save_condition:
            if self.config.save_root_path:
                save_path = os.path.join(
                    self.config.save_root_path, self.config.save_dir.get(stage, "")
                )
            else:
                save_path = os.path.join(
                    rt_config["log_dir"],
                    rt_config["name"],
                    self.config.save_dir.get(stage, ""),
                )
            if "current_step" in rt_config:
                save_file = os.path.join(
                    save_path, f"step_{str(rt_config['current_step'])}_predict.json"
                )
            else:
                save_file = os.path.join(save_path, "predict.json")
            logger.info(f"Save the {stage} predict data at {save_file}")
            with open(save_file, "w") as f:
                json.dump(predicts, f, indent=4, ensure_ascii=False)

    @property
    def without_ground_truth_stage(self) -> set:
        """there is not groud truth in the returned stage

        Returns:
            without_ground_truth_stage

        """
        return {"predict", "online"}

    def process(
        self,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
        save_condition: bool = False,
    ) -> Union[Dict, List]:
        """PostProcess entry

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            save_condition: if save_condition is True, will force save the predict on all stage except online

        Returns:
            the log_info(metrics) or the stage is "online" return the predicts

        """
        log_info = {}
        if stage not in self.without_ground_truth_stage:
            average_loss = self.average_loss(list_batch_outputs=list_batch_outputs)
            for name in average_loss:
                log_info[f"{self.loss_name_map(stage)}_{name}"] = average_loss[name]
        predicts = self.do_predict(stage, list_batch_outputs, origin_data, rt_config)
        if stage not in self.without_ground_truth_stage:
            log_info.update(
                self.do_calc_metrics(
                    predicts, stage, list_batch_outputs, origin_data, rt_config
                )
            )

        if stage == "online":
            return predicts
        if stage == "predict":
            if save_condition:
                self.do_save(
                    predicts,
                    stage,
                    list_batch_outputs,
                    origin_data,
                    rt_config,
                    save_condition=True,
                )
            return predicts
        else:
            if save_condition:
                self.do_save(
                    predicts,
                    stage,
                    list_batch_outputs,
                    origin_data,
                    rt_config,
                    save_condition=True,
                )
            else:
                self.do_save(
                    predicts,
                    stage,
                    list_batch_outputs,
                    origin_data,
                    rt_config,
                    save_condition=False,
                )
            return log_info

    def __call__(
        self, stage, list_batch_outputs, origin_data, rt_config, save_condition=False
    ):
        """the same as self.process"""
        return self.process(
            stage, list_batch_outputs, origin_data, rt_config, save_condition
        )


postprocessor_dir = os.path.dirname(__file__)
import_module_dir(postprocessor_dir, "dlk.data.postprocessor")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging

from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from pyecharts.charts import Bar
from pyecharts.options import InitOpts

from dlk.display import Display, DisplayConfigBase
from dlk.utils.register import register

logger = logging.getLogger(__name__)


@cregister("display", "summary")
class SummaryDisplayConfig(DisplayConfigBase):
    """the summary display"""

    title = StrField("Text Summary Task Demo", help="the title of the demo")
    help = StrField(
        "You can add the origin input to the `input area` and click `submit` to get the summary text of the input",
        help="the help infomation of the demo",
    )
    render_height = IntField(value=200, help="the display height")
    render_width = IntField(value=600, help="the display width")
    render_type = StrField(value="text", help="the render type of the display")
    output_head = StrField(value="Summary", help="the output head")

    class Input:
        input = StrField(
            value="text",
            options=["text"],
            help="the input values name for summary, the value is the input type",
        )

    input = NestField(value=Input, help="the input config")

    class Hold:
        target = StrField("", options=[""], help="the default target of summary")

    hold = NestField(value=Hold, help="the hold values")


@register("display", "summary")
class SummaryDisplay(Display):
    """text classification display"""

    def __init__(self, config: SummaryDisplayConfig):
        super(SummaryDisplay, self).__init__(config)

    def display(self, result):
        generate_summary = result["generated"][0]["generate"]
        return generate_summary



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import math
from typing import Dict

import torch.optim as optim
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts

from dlk.utils.register import register

from . import BaseScheduler, BaseSchedulerConfig

logger = logging.getLogger(__name__)


@cregister("scheduler", "cosine_restart")
class CosineRestartScheduleConfig(BaseSchedulerConfig):
    """
    the cosine restart scheduler
    """

    first_restart_step = IntField(
        value=-1,
        help="if the interval is `step`, the first restart step, if the interval is `epoch`, it means the first restart epoch",
    )
    mult_fact = IntField(
        value=1,
        minimum=0,
        help="the multiplier for the next restart interval, A factor increases after a restart. Default: 1.",
    )
    eta_min = FloatField(
        value=0.0, minimum=0.0, help="Minimum learning rate. Default: 0."
    )


@register("scheduler", "cosine_restart")
class CosineRestartSchedule(BaseScheduler):
    """CosineRestartSchedule"""

    def __init__(
        self, optimizer: optim.Optimizer, config: CosineRestartScheduleConfig, rt_config
    ):
        super(CosineRestartSchedule, self).__init__(optimizer, config, rt_config)
        self.config = config

    def get_scheduler(self) -> CosineAnnealingWarmRestarts:
        """return the initialized linear wramup then cos decay scheduler

        Returns:
            Schedule

        """
        return CosineAnnealingWarmRestarts(
            self.optimizer,
            T_0=self.config.first_restart_step,
            T_mult=self.config.mult_fact,
            eta_min=self.config.eta_min,
            last_epoch=-1,
        )



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
from collections import Counter
from typing import Dict, Iterable, List, Union

import pandas as pd

from dlk.utils.io import open


class Vocabulary(object):
    """generate vocab from tokens(token or Iterable tokens)
    you can dumps the object to dict and load from dict
    """

    def __init__(
        self, do_strip: bool = False, unknown: str = "", ignore: str = "", pad: str = ""
    ):
        self.word2idx = {}
        self.idx2word = {}
        self.do_strip = do_strip
        self.word_num = 0
        self.word_count = Counter()  # reserved
        self.unknown = unknown
        self.ignore = ignore
        self.pad = pad
        if ignore:
            self.word2idx[ignore] = -100
            self.idx2word[-100] = ignore
            self.word_count[ignore] += int(1e10)
        if pad:
            assert self.word_num == 0, f"The pad id must be 0"
            self.word_count[pad] += int(1e10)
            self.word2idx[pad] = self.word_num
            self.idx2word[self.word_num] = pad
            self.word_num += 1
        if unknown:
            self.word_count[unknown] += int(1e10)
            self.word2idx[unknown] = self.word_num
            self.idx2word[self.word_num] = unknown
            self.word_num += 1

    def dumps(self) -> Dict:
        """dumps the object to dict

        Returns:
            self.__dict__

        """
        return self.__dict__

    def dump(self, path: str):
        """dump the object to path

        Returns:
            self.__dict__

        """
        with open(path, "w", encoding="utf-8") as f:
            json.dump(self.__dict__, f)

    @classmethod
    def load(cls, attr: Dict):
        """load the object from dict

        Args:
            attr: self.__dict__

        Returns:
            initialized Vocabulary

        """
        vocab = cls()
        vocab.__dict__ = attr
        return vocab

    @classmethod
    def load_from_file(cls, path: str):
        """load the object from path

        Args:
            path: a json file

        Returns:
            initialized Vocabulary

        """
        with open(path, "r", encoding="utf-8") as f:
            attr = json.load(f)
        vocab = cls()
        vocab.__dict__ = attr
        id2word = vocab.idx2word
        vocab.idx2word = {int(i): id2word[i] for i in id2word}
        return vocab

    def __getitem__(self, index: int):
        """get the token by index

        Args:
            index: token index

        Returns:
            `word` which index is geven, if index is not out of range

        Raises:
            KeyError

        """
        return self.idx2word[int(index)]

    def auto_get_index(self, data: Union[str, List]):
        """get the index of word ∈data from this vocab

        Args:
            data: auto detection

        Returns:
            type the same as data

        """
        if isinstance(data, str):
            return self.get_index(data)
        elif isinstance(data, list):
            return [self.auto_get_index(subdata) for subdata in data]
        else:
            raise ValueError("Don't support the type of {}".format(data))

    def get_index(self, word: str) -> int:
        """get the index of word from this vocab

        Args:
            word: a single token

        Returns:
            index

        """
        if self.do_strip:
            word = word.strip()
        try:
            return self.word2idx[word]
        except:
            if self.unknown:
                return self.word2idx[self.unknown]
            else:
                raise KeyError("Unkown word: {}".format(word))

    def filter_rare(self, min_freq=1, most_common=-1):
        """filter the words which count is to small.

        min_freq and most_common can not set all

        Args:
            min_freq: minist frequency
            most_common: most common number, -1 means all

        Returns:
            None

        """
        self.word2idx = {}
        self.idx2word = {}
        assert (
            min_freq == 1 or most_common == -1
        ), "You should set the min_freq=1 or most_common=-1."
        if most_common != -1:
            for i, (token, freq) in enumerate(self.word_count.most_common(most_common)):
                self.word2idx[token] = i
                self.idx2word[i] = token
        else:
            index = 0
            for token in self.word_count:
                if self.word_count[token] >= min_freq:
                    self.word2idx[token] = index
                    self.idx2word[index] = token
                    index += 1
        return self

    def get_word(self, index: int) -> str:
        """get the word by index

        Args:
            index: word index

        Returns:
            word

        """

        try:
            return self.idx2word[int(index)]
        except:
            if index == -1:
                return "[unknown]"
            raise KeyError("Undefined index: {}".format(index))

    def add(self, word):
        """add one word to vocab

        Args:
            word: single word

        Returns:
            self

        """

        if not self.word_count[word]:
            self.word2idx[word] = self.word_num
            self.idx2word[self.word_num] = word
            self.word_num += 1
        self.word_count[word] += 1
        return self

    def auto_update(self, data: Union[str, Iterable]):
        """auto detect data type to update the vocab

        Args:
            data:  str| List[str] | Set[str] | List[List[str]]

        Returns:
            self

        """
        if isinstance(data, str):
            self.add(data)
        elif (
            isinstance(data, list)
            or isinstance(data, set)
            or isinstance(data, pd.Series)
        ):
            self.add_from_iter(data)
        else:
            raise ValueError("Don't support the type of {}".format(data))
        return self

    def __len__(self):
        """get the token num of vocab
        Returns:
            len(self.word2idx)

        """
        return len(self.word2idx)

    def add_from_iter(self, iterator):
        """add the tokens in iterator to vocab

        Args:
            iterator: List[str] | Set[str] | List[List[str]]

        Returns:
            self

        """
        for word in iterator:
            if isinstance(word, list) or isinstance(word, set):
                self.add_from_iter(word)
            elif isinstance(word, str):
                self.add(word)
            else:
                raise ValueError("Don't support the type of {}".format(word))
        return self



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Callable, Dict, List, Tuple

import torch

from dlk.utils.import_module import import_module_dir

imodel_dir = os.path.dirname(__file__)
import_module_dir(imodel_dir, "dlk.imodel")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
import uuid
from typing import Dict, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd
import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from tokenizers import Tokenizer

from dlk.data.postprocessor import BasePostProcessor, BasePostProcessorConfig
from dlk.utils.io import open
from dlk.utils.register import register
from dlk.utils.vocab import Vocabulary

logger = logging.getLogger(__name__)


@cregister("postprocessor", "span_relation")
class SpanRelationPostProcessorConfig(BasePostProcessorConfig):
    """span based relation extraction postprocessor"""

    ignore_position = BoolField(
        value=False, help="whether to ignore the position of the entity"
    )
    ignore_char = ListField(
        value=[],
        suggestions=[[" ", "(", ")", "[", "]", "-", ".", ",", ":", "'", '"']],
        help="ignore the provided char if these char is prefix or suffix of the entity",
    )
    entity_label_vocab = StrField(
        value=MISSING,
        help="the label vocab file path of entity, it should be the same as file in the preprocessor",
    )
    relation_label_vocab = StrField(
        value=MISSING,
        help="the label vocab file path of relation, it should be the same as file in the preprocessor",
    )
    tokenizer_path = StrField(
        value=MISSING, help="the tokenizer file path, is not effected by `meta_dir`"
    )

    class InputMap:
        entity_logits = StrField(value="entity_logits", help="the entity logits")
        relation_logits = StrField(value="relation_logits", help="the relation logits")
        index = StrField(value="_index", help="the index of the sample")

    input_map = NestField(
        value=InputMap,
        help="the input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )

    class OriginInputMap:
        uuid = StrField(value="uuid", help="the uuid or the id of the sample")
        sentence = StrField(value="sentence", help="the sentence of the sample")
        input_ids = StrField(value="input_ids", help="the input ids of the sample")
        entities_info = StrField(value="entities_info", help="the entities info")
        relations_info = StrField(value="relations_info", help="the relations info")
        offsets = StrField(value="offsets", help="the offsets of the tokens")
        special_tokens_mask = StrField(
            value="special_tokens_mask", help="the special tokens mask"
        )
        word_ids = StrField(value="word_ids", help="the word ids")

    origin_input_map = NestField(
        value=OriginInputMap,
        help="the origin input map of the processor, the key is the name of the processor needed key, the value is the provided data provided key",
    )
    unrelated_entity = BoolField(
        value=True, help="whether to save the unrelated entity(no pair relation)"
    )
    relation_groups = IntField(
        value=1,
        help="the relation groups, when set label_seperate==True in span_relation_relabel, the relation_groups maybe >1, but currently is not Tested",
    )
    sym = BoolField(
        value=True,
        help="whether the relation is sym, if sym==True, we can just calc the upper trim and set down trim as -100 to ignore",
    )
    ignore_labels = ListField(
        value=["O", "X", "S", "E"],
        help="the ignore labels, if the entity label in this list, we will ignore this entity",
    )
    ignore_relations = ListField(value=[], help="the ignore relations")


@register("postprocessor", "span_relation")
class SpanRelationPostProcessor(BasePostProcessor):
    """PostProcess for sequence labeling task"""

    def __init__(self, config: SpanRelationPostProcessorConfig):
        super(SpanRelationPostProcessor, self).__init__(config)
        self.config = config

        self.entity_label_vocab = Vocabulary.load_from_file(
            os.path.join(self.config.meta_dir, self.config.entity_label_vocab)
        )
        self.relation_label_vocab = Vocabulary.load_from_file(
            os.path.join(self.config.meta_dir, self.config.relation_label_vocab)
        )
        with open(self.config.tokenizer_path, "r", encoding="utf-8") as f:
            tokenizer_str = json.dumps(json.load(f))
        self.tokenizer = Tokenizer.from_str(tokenizer_str)

    def do_predict(
        self,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> List:
        """Process the model predict to human readable format

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            all predicts

        """
        predicts = []
        if self.config.origin_input_map.sentence not in origin_data:
            logger.error(
                f"{self.config.origin_input_map.sentence} not in the origin data"
            )
            raise PermissionError(
                f"{self.config.origin_input_map.sentence} must be provided"
            )
        if self.config.origin_input_map.uuid not in origin_data:
            logger.error(f"{self.config.origin_input_map.uuid} not in the origin data")
            raise PermissionError(
                f"{self.config.origin_input_map.uuid} must be provided"
            )
        predicts = []
        for outputs in list_batch_outputs:
            batch_entity_logits = outputs[self.config.input_map.entity_logits]
            batch_relation_logits = outputs[self.config.input_map.relation_logits]

            indexes = list(outputs[self.config.input_map.index])
            for entity_logits, relation_logits, index in zip(
                batch_entity_logits, batch_relation_logits, indexes
            ):
                one_ins_info = self._process4predict(
                    entity_logits, relation_logits, index, origin_data
                )
                predicts.append(one_ins_info)
        return predicts

    def _process4predict(
        self,
        entity_logits: torch.FloatTensor,
        relation_logits: torch.FloatTensor,
        index: int,
        origin_data: pd.DataFrame,
    ) -> Dict:
        """gather the predict and origin text and ground_truth_entities_info for predict

        Args:
            entity_logits: the predict entity span logits
            relation_logits: the predict relation logits
            index: the data index in origin_data
            origin_data: the origin pd.DataFrame

        Returns:
            >>> one_ins info
            >>> {
            >>>     "sentence": "...",
            >>>     "uuid": "..",
            >>>     "entities_info": [".."],
            >>>     "predict_entities_info": [".."],
            >>> }

        """

        def _get_entity_info(
            sub_tokens_index: List, offset_mapping: List, word_ids: List, label: str
        ) -> Dict:
            """gather sub_tokens to get the start and end

            Args:
                sub_tokens_index: the entity tokens index list
                offset_mapping: every token offset in text
                word_ids: every token in the index of words
                label: predict label

            Returns:
                entity_info

            """
            if not sub_tokens_index or not label:
                return {}
            start = offset_mapping[sub_tokens_index[0]][0]
            end = offset_mapping[sub_tokens_index[-1]][1]
            return {
                "start": start,
                "end": end,
                "labels": [label],
                "sub_token_start": sub_tokens_index[0],
                "sub_token_end": sub_tokens_index[1],
            }

        one_ins = {}
        origin_ins = origin_data.iloc[int(index)]
        one_ins["sentence"] = origin_ins[self.config.origin_input_map.sentence]
        one_ins["uuid"] = origin_ins[self.config.origin_input_map.uuid]
        one_ins["entities_info"] = origin_ins.get(
            self.config.origin_input_map.entities_info, []
        )
        one_ins["relations_info"] = origin_ins.get(
            self.config.origin_input_map.relations_info, []
        )

        word_ids = origin_ins[self.config.origin_input_map.word_ids]
        rel_token_len = len(word_ids)
        offset_mapping = origin_ins[self.config.origin_input_map.offsets][
            :rel_token_len
        ]

        predict_entities_id_info_map = {}
        entities_in_relations_id = set()
        predict_entity_ids = entity_logits.argmax(-1).cpu().numpy()

        # gather all predicted entities
        entity_set_d = (
            {}
        )  # set_d, set_e, set_t defined in https://arxiv.org/pdf/2010.13415.pdf Algorithm 1
        entity_cnt = 0
        for i in range(rel_token_len):
            for j in range(i, rel_token_len):
                if word_ids[i] is None or word_ids[j] is None:
                    continue
                predict_entity_id = predict_entity_ids[i][j]
                predict_entity = self.entity_label_vocab[predict_entity_id]
                if predict_entity in {
                    self.entity_label_vocab.pad,
                    self.entity_label_vocab.unknown,
                }:
                    continue
                else:
                    entity_info = _get_entity_info(
                        [i, j], offset_mapping, word_ids, predict_entity
                    )
                    if entity_info:
                        entity_id = str(uuid.uuid1())
                        predict_entities_id_info_map[entity_id] = entity_info
                        entity_set_d[i] = entity_set_d.get(i, [])
                        entity_set_d[i].append((i, j, entity_id))
                        entity_cnt += 1
            if entity_cnt > rel_token_len:
                # HACK: too many predict, maybe wrong
                break

        # all predicted relations entity tail pair
        entity_tail_pair_set_e = (
            set()
        )  # for each element (first_entity_tail_token_idx, second_entity_tail_token_idx, relation_idx)
        entity_tail_pair_set_e_with_relation_id = (
            set()
        )  # for each element (first_entity_tail_token_idx, second_entity_tail_token_idx, relation_idx, tail_relation_label_id)

        # all candidate relation set(only consider the entity head pair and the entity_set_d)
        candidate_relation_set_c = (
            set()
        )  # for each element (first_entity_info, second_entity_info, relation_idx, head_relation_label_id)
        candidate_cnt = 0
        for relation_idx in range(self.config.relation_groups):
            head_to_head_logits = relation_logits[relation_idx * 2]
            tail_to_tail_logits = relation_logits[relation_idx * 2 + 1]
            head_to_head_ids = head_to_head_logits.argmax(-1).cpu().numpy()
            tail_to_tail_ids = tail_to_tail_logits.argmax(-1).cpu().numpy()
            for i in range(rel_token_len):
                if candidate_cnt > 2 * rel_token_len:
                    # HACK: too many predict, maybe wrong
                    break
                for j in range(i if self.config.sym else 0, rel_token_len):
                    if word_ids[i] is None or word_ids[j] is None:
                        continue
                    predict_tail_id = tail_to_tail_ids[i][j]
                    predict_tail_relation = self.config.relation_label_vocab[
                        predict_tail_id
                    ]
                    if predict_tail_relation not in {
                        self.relation_label_vocab.pad,
                        self.relation_label_vocab.unknown,
                    }:
                        entity_tail_pair_set_e.add((i, j, relation_idx))
                        entity_tail_pair_set_e_with_relation_id.add(
                            (i, j, relation_idx, predict_tail_id)
                        )

                    predict_head_id = head_to_head_ids[i][j]
                    predict_head_relation = self.relation_label_vocab[predict_head_id]
                    if predict_head_relation not in {
                        self.relation_label_vocab.pad,
                        self.relation_label_vocab.unknown,
                    }:
                        for first_entity in entity_set_d.get(i, []):
                            for second_entity in entity_set_d.get(j, []):
                                candidate_relation_set_c.add(
                                    (
                                        first_entity,
                                        second_entity,
                                        relation_idx,
                                        predict_head_id,
                                    )
                                )
                                candidate_cnt += 1
        predict_relations_info = []
        for candidate_relation in candidate_relation_set_c:
            (
                first_entity,
                second_entity,
                relation_idx,
                predict_head_id,
            ) = candidate_relation
            if self.config.sym and first_entity[1] > second_entity[1]:
                first_entity, second_entity = second_entity, first_entity
            if (
                first_entity[1],
                second_entity[1],
                relation_idx,
            ) in entity_tail_pair_set_e:
                # HACK: if (first_entity[1], second_entity[1], relation_idx, predict_head_id) not in entity_tail_pair_set_e_with_relation_id, we can do better on it
                predict_label = self.relation_label_vocab[predict_head_id]

                if first_entity[2] not in entities_in_relations_id:
                    entities_in_relations_id.add(first_entity[2])
                if second_entity[2] not in entities_in_relations_id:
                    entities_in_relations_id.add(second_entity[2])
                predict_relation_info = {
                    "from": first_entity[2],
                    "to": second_entity[2],
                    "labels": [predict_label],
                }
                predict_relations_info.append(predict_relation_info)

        entity_ids = entities_in_relations_id
        if self.config.unrelated_entity:
            entity_ids = predict_entities_id_info_map.keys()
        predict_entities_info = []
        for entity_id in entity_ids:
            entity_info = predict_entities_id_info_map[entity_id]
            entity_info["entity_id"] = entity_id
            predict_entities_info.append(entity_info)

        one_ins["predict_entities_info"] = predict_entities_info
        one_ins["predict_relations_info"] = predict_relations_info
        return one_ins

    def do_calc_metrics(
        self,
        predicts: List,
        stage: str,
        list_batch_outputs: List[Dict],
        origin_data: pd.DataFrame,
        rt_config: Dict,
    ) -> Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns:
            the named scores, recall, precision, f1

        """
        real_name = self.loss_name_map(stage)
        entity_precision, entity_recall, entity_f1 = self._do_calc_entity_metrics(
            predicts, list_batch_outputs
        )
        (
            relation_precision,
            relation_recall,
            relation_f1,
        ) = self._do_calc_relation_metrics(predicts, list_batch_outputs)
        logger.info(
            f"{real_name:>8}_entity_precision: {entity_precision*100:.2f}, {real_name:>8}_recall: {entity_recall*100:.2f}, {real_name:>8}_f1: {entity_f1*100:.2f}"
        )
        logger.info(
            f"{real_name:>8}_relation_precision: {relation_precision*100:.2f}, {real_name:>8}_recall: {relation_recall*100:.2f}, {real_name:>8}_f1: {relation_f1*100:.2f}"
        )
        return {
            f"{real_name}_ent_p": entity_precision * 100,
            f"{real_name}_ent_r": entity_recall * 100,
            f"{real_name}_ent_f1": entity_f1 * 100,
            f"{real_name}_rel_p": relation_precision * 100,
            f"{real_name}_rel_r": relation_recall * 100,
            f"{real_name}_rel_f1": relation_f1 * 100,
        }

    def _do_calc_relation_metrics(self, predicts: List, list_batch_outputs: List[Dict]):
        """calc relation related metrics
        Returns:
            relation related metrics

        """

        def _group_relations_info(
            relations_info: List[Dict], entities_info: List[Dict]
        ) -> Dict[str, Set[Tuple]]:
            """flat the relations info by relations info dict and entities info dict

            Args:
                relations_info (List[Dict]): like
                        [
                            {
                                "labels": [
                                    "belong_to"
                                ],
                                "from": "bd7a2928-52a8-11ed-8b5c-18c04d299e80", # id of entity
                                "to": "bd798da3-52a8-11ed-801c-18c04d299e80",
                            }
                        ]
                entities_info (List[Dict]): like
                        [
                            {
                                "entity_id": "bd798da3-52a8-11ed-801c-18c04d299e80",
                                "start_sub_token":1,
                                "end_sub_token": 3,
                                "labels": [
                                    "Brand"
                                ]
                            },
                            {
                                "entity_id": "bd7a2928-52a8-11ed-8b5c-18c04d299e80",
                                "start_sub_token":5,
                                "end_sub_token": 9,
                                "labels": [
                                    "Product"
                                ]
                            }
                        ],

            Returns:
                relations type relation info set pair, like
                    {
                        "belong_to": {
                            (1, 3, 5, 9),
                            ...
                        }
                    }

            Explanation:
                in the returns example,
                    1 ==> from_entity.sub_token_start
                    3 ==> from_entity.sub_token_end
                    5 ==> to_entity.sub_token_start
                    9 ==> to_entity.sub_token_end
            belong_to ==> the relation type
            """

            def _norm_entity(entity):
                """norm the entity start and end by the ignore_char

                Args:
                    entity: one entity is like:
                        {
                            "start": .,
                            "end": .,
                            "sub_token_start": .,
                            "sub_token_end": .,
                        }
                Returns: norm entity
                """
                start_position, end_position = entity["start"], entity["end"]
                while start_position < end_position:
                    if text[start_position] in self.config.ignore_char:
                        start_position += 1
                    else:
                        break
                while start_position < end_position:
                    if text[end_position - 1] in self.config.ignore_char:
                        end_position -= 1
                    else:
                        break
                if (
                    start_position == end_position
                ):  # if the entity after remove ignore char be null, we set it to origin
                    return entity["start"], entity["end"]
                return start_position, end_position

            entities_id_info_map = {}
            for entity_info in entities_info:
                entities_id_info_map[entity_info["entity_id"]] = entity_info

            flat_relations = {}
            for relation_info in relations_info:
                from_entity = entities_id_info_map[relation_info["from"]]
                to_entity = entities_id_info_map[relation_info["to"]]
                relation_type = relation_info["labels"][
                    0
                ]  # NOTE: you can also change the relation type format for different label format
                if relation_type not in flat_relations:
                    flat_relations[relation_type] = set()
                from_start, from_end = _norm_entity(from_entity)
                to_start, to_end = _norm_entity(to_entity)

                # flat_relations[relation_type].add((from_entity['sub_token_start'], from_entity['sub_token_end'], to_entity['sub_token_start'], to_entity['sub_token_end']))
                flat_relations[relation_type].add(
                    (from_start, from_end, to_start, to_end)
                )
            return flat_relations

        relation_match_info = {}
        for predict in predicts:
            text = predict["sentence"]
            predict_relations = _group_relations_info(
                predict["predict_relations_info"], predict["predict_entities_info"]
            )
            ground_truth_relations = _group_relations_info(
                predict["relations_info"], predict["entities_info"]
            )
            for key in set(ground_truth_relations.keys()).union(
                set(predict_relations.keys())
            ):
                if key not in relation_match_info:
                    relation_match_info[key] = {
                        "match": 0,
                        "miss": 0,
                        "wrong": 0,
                    }
            for key in ground_truth_relations:
                for ground_truth_relation in ground_truth_relations[key]:
                    if ground_truth_relation in predict_relations.get(key, {}):
                        predict_relations[key].remove(ground_truth_relation)
                        relation_match_info[key]["match"] += 1
                    elif self.config.sym and (
                        ground_truth_relation[2],
                        ground_truth_relation[3],
                        ground_truth_relation[0],
                        ground_truth_relation[1],
                    ) in predict_relations.get(key, {}):
                        predict_relations[key].remove(
                            (
                                ground_truth_relation[2],
                                ground_truth_relation[3],
                                ground_truth_relation[0],
                                ground_truth_relation[1],
                            )
                        )
                        relation_match_info[key]["match"] += 1
                    else:
                        relation_match_info[key]["miss"] += 1
            for key in predict_relations:
                for wrong in predict_relations[key]:
                    relation_match_info[key]["wrong"] += 1
        all_tp, all_fn, all_fp = 0, 0, 0

        def _care_div(a, b):
            """return a/b or 0.0 if b == 0"""
            if b == 0:
                return 0.0
            return a / b

        for key in relation_match_info:
            tp = relation_match_info[key]["match"]
            fn = relation_match_info[key]["miss"]
            fp = relation_match_info[key]["wrong"]
            precision = _care_div(tp, tp + fp)
            recall = _care_div(tp, tp + fn)
            f1 = 2 * _care_div(precision * recall, precision + recall)

            all_tp += tp
            all_fn += fn
            all_fp += fp
            logger.info(
                f"For {'relation':16} 「{key[:16]:16}」, the precision={precision*100 :.2f}%, the recall={recall*100:.2f}%, f1={f1*100:.2f}%"
            )

        precision = _care_div(all_tp, all_tp + all_fp)
        recall = _care_div(all_tp, all_tp + all_fn)
        f1 = 2 * _care_div(precision * recall, precision + recall)
        return precision, recall, f1

    def _do_calc_entity_metrics(self, predicts: List, list_batch_outputs: List[Dict]):
        """calc entity related metrics
        Returns:
            entity related metrics

        """

        def _group_entities_info(entities_info: List[Dict], text: str) -> Dict:
            """gather the same labeled entity to the same list

            Args:
                entities_info:
                    >>> [
                    >>>     {
                    >>>         "start": start1,
                    >>>         "end": end1,
                    >>>         "labels": ["label_1"]
                    >>>     },
                    >>>     {
                    >>>         "start": start2,
                    >>>         "end": end2,
                    >>>         "labels": ["label_2"]
                    >>>     },....
                    >>> ]
                text: be labeled text

            Returns:
                >>> { "label_1" [text[start1:end1]], "label_2": [text[start_2: end_2]]...}

            """
            info = {}
            for item in entities_info:
                label = item["labels"][0]
                if label not in info:
                    info[label] = []
                start_position, end_position = item["start"], item["end"]
                while start_position < end_position:
                    if text[start_position] in self.config.ignore_char:
                        start_position += 1
                    else:
                        break
                while start_position < end_position:
                    if text[end_position - 1] in self.config.ignore_char:
                        end_position -= 1
                    else:
                        break
                if (
                    start_position == end_position
                ):  # if the entity after remove ignore char be null, we set it to origin
                    start_position, end_position = item["start"], item["end"]

                if self.config.ignore_position:
                    info[label].append(text[item["start"] : item["end"]].strip())
                else:
                    info[label].append((start_position, end_position))
            return info

        def _calc_entity_score(predict_list: List, ground_truth_list: List):
            """use predict_list and ground_truth_list to calc scores

            Args:
                predict_list: list of predict
                ground_truth_list: list of ground_truth

            Returns:
                precision, recall, f1

            """
            category_tp = {}
            category_fp = {}
            category_fn = {}

            def _care_div(a, b):
                """return a/b or 0.0 if b == 0"""
                if b == 0:
                    return 0.0
                return a / b

            def _calc_num(_pred: List, _ground_truth: List):
                """calc tp, fn, fp

                Args:
                    pred: pred list
                    ground_truth: groud truth list

                Returns:
                    tp, fn, fp

                """
                num_p = len(_pred)
                num_t = len(_ground_truth)
                truth = 0
                for p in _pred:
                    if p in _ground_truth:
                        truth += 1
                return truth, num_t - truth, num_p - truth

            for predict, ground_truth in zip(predict_list, ground_truth_list):
                keys = set(list(predict.keys()) + list(ground_truth.keys()))
                for key in keys:
                    tp, fn, fp = _calc_num(
                        predict.get(key, []), ground_truth.get(key, [])
                    )
                    category_tp[key] = category_tp.get(key, 0) + tp
                    category_fn[key] = category_fn.get(key, 0) + fn
                    category_fp[key] = category_fp.get(key, 0) + fp

            all_tp, all_fn, all_fp = 0, 0, 0
            for key in category_tp:
                tp, fn, fp = category_tp[key], category_fn[key], category_fp[key]
                all_tp += tp
                all_fn += fn
                all_fp += fp
                precision = _care_div(tp, tp + fp)
                recall = _care_div(tp, tp + fn)
                f1 = _care_div(2 * precision * recall, precision + recall)
                logger.info(
                    f"For {'entity':16} 「{key[:16]:16}」, the precision={precision*100 :.2f}%, the recall={recall*100:.2f}%, f1={f1*100:.2f}%"
                )
            precision = _care_div(all_tp, all_tp + all_fp)
            recall = _care_div(all_tp, all_tp + all_fn)
            f1 = _care_div(2 * precision * recall, precision + recall)
            return precision, recall, f1

        all_predicts = []
        all_ground_truths = []
        for predict in predicts:
            text = predict["sentence"]
            predict_ins = _group_entities_info(predict["predict_entities_info"], text)
            ground_truth_ins = _group_entities_info(predict["entities_info"], text)
            all_predicts.append(predict_ins)
            all_ground_truths.append(ground_truth_ins)

        entity_precision, entity_recall, entity_f1 = _calc_entity_score(
            all_predicts, all_ground_truths
        )
        return entity_precision, entity_recall, entity_f1



# Copyright the author(s) of DLK.

# There are many code copy from fairseq.
# Copyright (c) Facebook, Inc. and its affiliates.


import math
from typing import Dict, List, Optional

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch import Tensor

from dlk.utils.register import register

from . import Sample


@cregister("token_sample", "sampling")
class SamplingConfig(Base):
    sampling_topp = FloatField(value=0, help="the top-p sampling")
    sampling_topk = IntField(
        value=0,
        help="the top-k sampling, if all sampling_topp and sampling_topk are set will use topp",
    )


@register("token_sample", "sampling")
class Sampling(Sample):
    sampling_topk: int
    sampling_topp: float

    def __init__(self, tgt_dict, config: SamplingConfig):
        super().__init__(tgt_dict)
        self.sampling_topk = config.sampling_topk
        self.sampling_topp = config.sampling_topp

    def _sample_topp(self, lprobs):
        """Sample among the smallest set of elements whose cumulative probability mass exceeds p.

        See `"The Curious Case of Neural Text Degeneration"
        (Holtzman et al., 2019) <https://arxiv.org/abs/1904.09751>`_.

        Args:
            lprobs: (bsz x input_beam_size x vocab_size)
                the model's log-probabilities over the vocabulary at the current step

        Return: A tuple of (trimed_probs, truncated_indices) where:
            trimed_probs: (bsz x input_beam_size x ?)
                the model's probabilities over the elements selected to sample from. The
                width of the third dimension is determined by top-P.
            truncated_indices: (bsz x input_beam_size x ?)
                the indices of the chosen elements.
        """
        probs = lprobs.exp_()

        # sort the last dimension (vocab dimension) in descending order
        sorted_probs, sorted_indices = probs.sort(descending=True)

        # compute a mask to indicate the words to be included in the top-P set.
        cumsum_probs = sorted_probs.cumsum(dim=2)
        mask = cumsum_probs.lt(self.sampling_topp)

        # note that mask was computed by 'lt'. One more word needs to be included
        # so that the cumulative probability mass can exceed p.
        cumsum_mask = mask.cumsum(dim=2)
        last_included = cumsum_mask[:, :, -1:]
        last_included.clamp_(0, mask.size()[2] - 1)
        mask = mask.scatter_(2, last_included, 1)

        # truncate unnecessary dims.
        max_dim = last_included.max()
        truncated_mask = mask[:, :, : max_dim + 1]
        truncated_probs = sorted_probs[:, :, : max_dim + 1]
        truncated_indices = sorted_indices[:, :, : max_dim + 1]

        # trim the words that are not in top-P by setting their probabilities
        # to 0, so that they would not be sampled later.
        trim_mask = ~truncated_mask
        trimed_probs = truncated_probs.masked_fill_(trim_mask, 0)
        return trimed_probs, truncated_indices

    def step(
        self,
        step: int,
        lprobs,
        scores,
        prev_output_tokens: Optional[Tensor] = None,
        original_batch_idxs: Optional[Tensor] = None,
    ):
        bsz, beam_size, vocab_size = lprobs.size()

        if step == 0:
            # at the first step all hypotheses are equally likely, so use
            # only the first beam
            lprobs = lprobs[:, ::beam_size, :].contiguous()

        if self.sampling_topp > 0:
            # only sample from the smallest set of words whose cumulative probability mass exceeds p
            probs, top_indices = self._sample_topp(lprobs)
        elif self.sampling_topk > 0:
            # only sample from top-k candidates
            lprobs, top_indices = lprobs.topk(self.sampling_topk)
            probs = lprobs.exp_()
        else:
            probs = lprobs.exp_()

            # dummy data to be consistent with true branch for type check
            top_indices = torch.empty(0).to(probs)
        # sample
        if step == 0:
            indices_buf = torch.multinomial(
                probs.view(bsz, -1),
                beam_size,
                replacement=True,
            ).view(bsz, beam_size)
        else:
            indices_buf = torch.multinomial(
                probs.view(bsz * beam_size, -1),
                1,
                replacement=True,
            ).view(bsz, beam_size)

        if step == 0:
            # expand to beam size
            probs = probs.expand(bsz, beam_size, -1)

        # gather scores
        scores_buf = torch.gather(probs, dim=2, index=indices_buf.unsqueeze(-1))
        scores_buf = scores_buf.log_().view(bsz, -1)

        # remap indices if using top-k or top-P sampling
        if self.sampling_topk > 0 or self.sampling_topp > 0:
            indices_buf = torch.gather(
                top_indices.expand(bsz, beam_size, -1),
                dim=2,
                index=indices_buf.unsqueeze(-1),
            ).squeeze(2)

        if step == 0:
            beams_buf = indices_buf.new_zeros(bsz, beam_size)
        else:
            beams_buf = torch.arange(0, beam_size).to(indices_buf).repeat(bsz, 1)
            # make scores cumulative
            scores_buf.add_(
                torch.gather(scores[:, :, step - 1], dim=1, index=beams_buf)
            )

        return scores_buf, indices_buf, beams_buf



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, Union

import lightning.pytorch as pl
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from lightning.pytorch.callbacks import ProgressBar, RichProgressBar, TQDMProgressBar
from lightning.pytorch.utilities.rank_zero import rank_zero_warn

from dlk.utils.register import register


@cregister("callback", "prograss_bar")
class ProgressBarCallbackConfig(Base):
    """callback for the prograss bar"""

    type_bar = StrField(
        value="tqdm",
        options=["tqdm", "rich"],
        help="""the type of the prograss bar, ``tqdm`` or ``rich``. Defaults to ``tqdm``.""",
    )
    refresh_rate = IntField(
        value=1,
        minimum=0,
        help="Determines at which rate (in number of batches) the progress bars get updated. Set it to ``0`` to disable the display. Defaults to ``1``.",
    )
    leave = BoolField(
        value=False,
        help="Leaves the finished progress bar in the terminal at the end of the epoch. Defaults to ``False``.",
    )
    theme = StrField(
        value="default",
        options=["default"],
        help="When the `type_bar` is `rich`. Contains styles used to stylize the progress bar. Currently only support `default`.",
    )
    process_position = IntField(
        value=0,
        help="Set this to a value greater than ``0`` to offset the progress bars by this many lines. This is useful when you have progress bars defined elsewhere and want to show all of them together.",
    )


class NewRichProgressBar(RichProgressBar):
    """docstring for  ProgressBar"""

    def __init__(self, **config):
        super(NewRichProgressBar, self).__init__(**config)

    def get_metrics(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> Dict[str, Union[int, str, float, Dict[str, float]]]:
        r"""Combines progress bar metrics collected from the trainer with standard metrics from
        get_standard_metrics. Implement this to override the items displayed in the progress bar.

        Here is an example of how to override the defaults:

        .. code-block:: python

            def get_metrics(self, trainer, model):
                # don't show the version number
                items = super().get_metrics(trainer, model)
                items.pop("v_num", None)
                return items

        Return:
            Dictionary with the items to be displayed in the progress bar.
        """
        standard_metrics = {}  # NOTE: remove the `v_num` get_standard_metrics(trainer)
        pbar_metrics = trainer.progress_bar_metrics
        duplicates = list(standard_metrics.keys() & pbar_metrics.keys())
        if duplicates:
            rank_zero_warn(
                f"The progress bar already tracks a metric with the name(s) '{', '.join(duplicates)}' and"
                f" `self.log('{duplicates[0]}', ..., prog_bar=True)` will overwrite this value. "
                " If this is undesired, change the name or override `get_metrics()` in the progress bar callback.",
            )

        return {**standard_metrics, **pbar_metrics}


class NewTQDMProgressBar(TQDMProgressBar):
    """docstring for  ProgressBar"""

    def __init__(self, **config):
        super(NewTQDMProgressBar, self).__init__(**config)

    def get_metrics(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> Dict[str, Union[int, str, float, Dict[str, float]]]:
        r"""Combines progress bar metrics collected from the trainer with standard metrics from
        get_standard_metrics. Implement this to override the items displayed in the progress bar.

        Here is an example of how to override the defaults:

        .. code-block:: python

            def get_metrics(self, trainer, model):
                # don't show the version number
                items = super().get_metrics(trainer, model)
                items.pop("v_num", None)
                return items

        Return:
            Dictionary with the items to be displayed in the progress bar.
        """
        standard_metrics = {}  # NOTE: remove the `v_num` get_standard_metrics(trainer)
        pbar_metrics = trainer.progress_bar_metrics
        duplicates = list(standard_metrics.keys() & pbar_metrics.keys())
        if duplicates:
            rank_zero_warn(
                f"The progress bar already tracks a metric with the name(s) '{', '.join(duplicates)}' and"
                f" `self.log('{duplicates[0]}', ..., prog_bar=True)` will overwrite this value. "
                " If this is undesired, change the name or override `get_metrics()` in the progress bar callback.",
            )

        return {**standard_metrics, **pbar_metrics}


@register("callback", "prograss_bar")
class RichProgressBarCallback(object):
    """prograss bar callback for lightning"""

    def __init__(self, config: ProgressBarCallbackConfig):
        super().__init__()
        self.config = config

    def __call__(self, rt_config: Dict) -> ProgressBar:
        """return LearningRateMonitor object

        Args:
            rt_config: runtime config, include save_dir, and the checkpoint path name

        Returns:
            RichProgressBar object

        """
        if self.config.type_bar == "tqdm":
            return NewTQDMProgressBar(
                refresh_rate=self.config.refresh_rate,
                process_position=self.config.process_position,
            )
        else:
            assert self.config.type_bar == "rich"
            theme = None
            assert self.config.theme == "default"
            from lightning.pytorch.callbacks.progress.rich_progress import (
                RichProgressBarTheme,
            )

            theme = RichProgressBarTheme()
            return NewRichProgressBar(
                refresh_rate=self.config.refresh_rate,
                leave=self.config.leave,
                theme=theme,
            )



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle as pkl
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from pyecharts.charts import Bar
from pyecharts.options import InitOpts

from dlk.display.txt_cls import (
    TextClassificationDisplay,
    TextClassificationDisplayConfig,
)
from dlk.utils.register import register

logger = logging.getLogger(__name__)


@cregister("display", "img_cls")
class ImageClassificationDisplayConfig(TextClassificationDisplayConfig):
    """the image classification display"""

    title = StrField("Image Classification Task Demo", help="the title of the demo")
    help = StrField(
        "You can upload the image to the `image upload` and click `submit` to get the class info of this sentence",
        help="the help infomation of the demo",
    )
    output_head = StrField(value="Image Class", help="the output head")

    class Input:
        image = StrField(
            "image",
            options=["image"],
            help="the input values name for classification, the value is the type",
        )

    input = NestField(value=Input, help="the input config")


@register("display", "img_cls")
class ImageClassificationDisplay(TextClassificationDisplay):
    """image classification display"""

    def __init__(self, config: ImageClassificationDisplayConfig):
        super(ImageClassificationDisplay, self).__init__(config)
        self.config = config



# Copyright the author(s) of DLK.

# There are many code copied from fairseq
# Copyright (c) Facebook, Inc. and its affiliates.


from typing import Dict

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from tokenizers import Tokenizer

from dlk.nn.base_module import BaseModel
from dlk.nn.token_gen_base import SpecialVocab, TokenGenBase, TokenGenBaseConfig
from dlk.utils.ngram_repeat_block import NGramRepeatBlock
from dlk.utils.register import register, register_module_name


@cregister("model", "token_enc_dec")
class TokenEncDecModelConfig(TokenGenBaseConfig):
    src_tokenizer = StrField(
        value=None,
        additions=[None],
        help="the source tokenizer file, if share_embedding is True, src_tokenizer should be None",
    )
    model_type = StrField(
        value="token_enc_dec",
        options=["dec_only", "token_enc_dec", "media_enc_token_dec"],
        help="the model type",
    )
    add_bos_token = BoolField(
        value=False,
        help="whether to add the begin-of-sentence token mannually, the most case is False, because the tokenizer will add it automatically",
    )

    src_embedding_dim = IntField(
        value=None,
        additions=[None],
        help="the target embedding dim, if set to None, will use the `tgt_embedding_dim`",
    )
    decoder_hidden_size = IntField(
        value=None,
        additions=[None],
        help="the hidden size of the decoder, if set to None, will use the `tgt_embedding_dim`",
    )
    max_len_a_ratio = FloatField(
        value=0.0,
        help="the max length ratio(base src), the real max_len = min(max_len_a_ratio * src_len + max_len_b, max_len)",
    )
    max_len_b = IntField(
        value=5,
        help="the targe max length, the real max_len = min(max_len_a_ratio * src_len + max_len_b, max_len)",
    )
    max_len = IntField(
        value=100,
        help="the maximum length, the real max_len = min(max_len_a_ratio * src_len + max_len_b, max_len)",
    )
    match_source_len = BoolField(
        value=False, help="outputs should match the sourcelength (default: False)"
    )


@register("model", "token_enc_dec")
class TokenEncDecModel(TokenGenBase):
    def __init__(self, config: TokenEncDecModelConfig, checkpoint):
        """Generates translations of a given source sentence.
        Args:
        """
        super().__init__(config, checkpoint)
        self.config = config
        if self.config.src_embedding_dim is None:
            self.config.src_embedding_dim = self.config.tgt_embedding_dim

        if config.share_embedding:
            self.src_vocab_size = self.tgt_vocab_size
        else:
            self.src_vocab_size = Tokenizer.from_file(
                config.src_tokenizer
            ).get_vocab_size()

            self.src_embedding = nn.Embedding(
                self.src_vocab_size, self.config.src_embedding_dim
            )

        if self.config.src_embedding_dim is None:
            self.config.src_embedding_dim = self.config.tgt_embedding_dim
        if self.config.decoder_hidden_size is None:
            self.config.decoder_hidden_size = self.config.tgt_embedding_dim

    def training_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do training for one batch

        Args:
            inputs: one mini-batch inputs

        Returns:
            the training outputs

        """
        decoder_input_ids = inputs[self.config.input_map.decoder_input_ids]
        bsz, _ = decoder_input_ids.shape
        if self.config.add_bos_token:
            decoder_input_ids = torch.concat(
                [
                    torch.LongTensor([self.bos] * bsz)
                    .view(-1, 1)
                    .to(decoder_input_ids),
                    decoder_input_ids,
                ],
                -1,
            )
            inputs[self.config.input_map.decoder_input_ids] = decoder_input_ids

        decoder_target_ids = torch.concat(
            [
                decoder_input_ids[:, 1:],
                torch.LongTensor([self.eos] * bsz).view(bsz, -1).to(decoder_input_ids),
            ],
            -1,
        )
        if self.config.share_embedding:
            src_embedding = (
                self.embedding.forward(inputs[self.config.input_map.encoder_input_ids])
                * self.embedding_scale
            )
        else:
            src_embedding = (
                self.src_embedding.forward(
                    inputs[self.config.input_map.encoder_input_ids]
                )
                * self.embedding_scale
            )
        inputs[self.encoder.config.input_map.encoder_input_embedding] = src_embedding
        encoder_outs = self.encoder.training_step(inputs)

        if self.config.share_embedding:
            decoder_embedding = (
                self.embedding.forward(
                    encoder_outs[self.config.input_map.decoder_input_ids]
                )
                * self.embedding_scale
            )
        else:
            decoder_embedding = (
                self.tgt_embedding.forward(
                    encoder_outs[self.config.input_map.decoder_input_ids]
                )
                * self.embedding_scale
            )
        encoder_outs[
            self.decoder.config.input_map.decoder_input_embedding
        ] = decoder_embedding
        decoder_outs = self.decoder.training_step(encoder_outs)

        decoder_output_embedding = decoder_outs[
            self.decoder.config.output_map.decoder_output_embedding
        ]
        logits = self.lm_head(decoder_output_embedding)
        decoder_outs[self.config.output_map.logits] = logits
        decoder_outs[self.config.output_map.decoder_target_ids] = decoder_target_ids
        return decoder_outs

    @torch.no_grad()
    def generate(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Generate tokens.

        Args:
            inputs: one mini-batch inputs
        """
        decoder_past_cache = None
        src_tokens = inputs[self.config.input_map.encoder_input_ids]
        inputs[self.config.output_map.decoder_target_ids] = inputs[
            self.config.input_map.decoder_input_ids
        ]
        # length of the source text being the character length except EndOfSentence and pad
        src_lengths = (
            (src_tokens.ne(self.eos) & src_tokens.ne(self.pad)).long().sum(dim=1)
        )

        # bsz: total number of sentences in beam
        # Note that src_tokens may have more than 2 dimensions (i.e. audio features)
        bsz, src_len = src_tokens.size()[:2]
        beam_size: int = self.config.beam_size

        if (
            "constraints" in inputs is not None
            and not self.token_sample.supports_constraints
        ):
            raise NotImplementedError(
                "Target-side constraints were provided, but token_sample method doesn't support them"
            )

        # Initialize constraints, when active
        self.token_sample.init_constraints(inputs.get("constraints", None), beam_size)

        max_len: int = -1
        if self.config.match_source_len:
            max_len = src_lengths.max().item()
        else:
            max_len = min(
                int(self.config.max_len_a_ratio * src_len + self.config.max_len_b),
                self.config.max_len - 1,
            )
        assert (
            self.config.min_len <= max_len
        ), "min_len cannot be larger than max_len, please adjust these!"

        if self.config.share_embedding:
            src_embedding = self.embedding.forward(src_tokens) * self.embedding_scale
        else:
            src_embedding = (
                self.src_embedding.forward(src_tokens) * self.embedding_scale
            )

        inputs[self.encoder.config.input_map.encoder_input_embedding] = src_embedding

        encoder_outs = self.encoder.forward(inputs)

        # placeholder of indices for bsz * beam_size to hold tokens and accumulative scores
        new_order = torch.arange(bsz).view(-1, 1).repeat(1, beam_size).view(-1)
        new_order = new_order.to(src_tokens.device).long()

        encoder_outs = self.encoder.reorder_encoder_out(encoder_outs, new_order)
        # ensure encoder_outs is a List.
        assert encoder_outs is not None

        inputs[self.config.output_map.generated] = self._generate_tokens(
            bsz, beam_size, max_len, encoder_outs, src_lengths
        )
        return inputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List

import torch
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.nn.base_module import BaseModel
from dlk.utils.register import register, register_module_name


@cregister("model", "basic")
class BasicModelConfig(Base):
    """
    The Basic Model, include the embedding, encoder, decoder
    """

    submodule = SubModule(
        value={},
        suggestions=[
            "embedding",
            "encoder",
            "decoder",
            "initmethod",
        ],
        help="submodules for basic model",
    )


@register("model", "basic")
class BasicModel(BaseModel):
    """Basic encode decode Model"""

    def __init__(self, config: BasicModelConfig, checkpoint):
        super().__init__()

        embedding_configs = config._get_modules("embedding")
        if len(embedding_configs) == 0:
            self.embedding = register.get("embedding", "identity")()
        else:
            assert len(embedding_configs) == 1
            self.embedding = register.get(
                "embedding", register_module_name(embedding_configs[0]._module_name)
            )(embedding_configs[0])

        encoder_configs = config._get_modules("encoder")
        if len(encoder_configs) == 0:
            self.encoder = register.get("encoder", "identity")()
        else:
            assert len(encoder_configs) == 1
            self.encoder = register.get(
                "encoder", register_module_name(encoder_configs[0]._module_name)
            )(encoder_configs[0])

        decoder_configs = config._get_modules("decoder")
        if len(decoder_configs) == 0:
            self.decoder = register.get("decoder", "identity")()
        else:
            assert len(decoder_configs) == 1
            self.decoder = register.get(
                "decoder", register_module_name(decoder_configs[0]._module_name)
            )(decoder_configs[0])

        if not checkpoint:
            init_method_configs = config._get_modules("initmethod")
            if len(init_method_configs) == 0:
                init_method = register.get("initmethod", "default")()
            else:
                assert len(init_method_configs) == 1
                init_method_config = init_method_configs[0]
                init_method = register.get(
                    "initmethod", register_module_name(init_method_config._module_name)
                )(init_method_config)
            self.embedding.init_weight(init_method)
            self.encoder.init_weight(init_method)
            self.decoder.init_weight(init_method)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns: the outputs

        """
        embedding_outputs = self.embedding(inputs)
        encode_outputs = self.encoder(embedding_outputs)
        decode_outputs = self.decoder(encode_outputs)
        return decode_outputs

    def predict_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do predict for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: the predicts outputs

        """
        embedding_outputs = self.embedding.predict_step(inputs)
        encode_outputs = self.encoder.predict_step(embedding_outputs)
        decode_outputs = self.decoder.predict_step(encode_outputs)
        return decode_outputs

    def training_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do training for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: the training outputs

        """
        embedding_outputs = self.embedding.training_step(inputs)
        encode_outputs = self.encoder.training_step(embedding_outputs)
        decode_outputs = self.decoder.training_step(encode_outputs)
        return decode_outputs

    def validation_step(
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """do validation for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: the validation outputs

        """
        embedding_outputs = self.embedding.validation_step(inputs)
        encode_outputs = self.encoder.validation_step(embedding_outputs)
        decode_outputs = self.decoder.validation_step(encode_outputs)
        return decode_outputs

    def test_step(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """do test for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: the test outputs

        """
        embedding_outputs = self.embedding.test_step(inputs)
        encode_outputs = self.encoder.test_step(embedding_outputs)
        decode_outputs = self.decoder.test_step(encode_outputs)
        return decode_outputs



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging

from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)

from dlk.display.summary import DisplayConfigBase, SummaryDisplay, SummaryDisplayConfig
from dlk.utils.register import register

logger = logging.getLogger(__name__)


@cregister("display", "img_caption")
class ImageCaptionDisplayConfig(DisplayConfigBase):
    """the image caption display"""

    title = StrField("Image Caption Task Demo", help="the title of the demo")
    help = StrField(
        "You can add the image to the `image area` and click `submit` to get the caption text of the image",
        help="the help infomation of the demo",
    )
    render_type = StrField(value="text", help="the output type of is text")
    output_head = StrField(value="Image Caption", help="the output head")

    class Input:
        image = StrField(
            value="image",
            help="the input name for img_caption, the value is the input type",
        )

    input = NestField(value=Input, help="the input config")

    class Hold:
        target = StrField("", options=[""], help="the default target of img_caption")

    hold = NestField(value=Hold, help="the hold values")


@register("display", "img_caption")
class ImageCaptionDisplay(SummaryDisplay):
    """text classification display"""

    def __init__(self, config: ImageCaptionDisplayConfig):
        super(ImageCaptionDisplay, self).__init__(config)



# Copyright the author(s) of DLK.
# Copyright spark-nlp-display.
#
# This original source code is copied from spark-nlp-display and licensed under the Apache license.

from intc import Base, DictField, IntField, StrField, dataclass


@dataclass
class SpanAnnotation(Base):
    annotatorType = StrField(
        value="",
        help="The type of the output of the annotator. Possible values are ``DOCUMENT, TOKEN, WORDPIECE, WORD_EMBEDDINGS, SENTENCE_EMBEDDINGS, CATEGORY, DATE, ENTITY, SENTIMENT, POS, CHUNK, NAMED_ENTITY, NEGEX, DEPENDENCY, LABELED_DEPENDENCY, LANGUAGE, KEYWORD, DUMMY``.",
    )
    begin = IntField(
        value=0, help="The index of the first character under this annotation."
    )
    end = IntField(
        value=0, help="The index of the last character under this annotation."
    )
    result = StrField(value="", help="The resulting string of the annotation.")
    metadata = DictField(value={}, help="Associated metadata for this annotation")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import math
from typing import Dict

import torch.optim as optim
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch.optim.lr_scheduler import LambdaLR

from dlk.utils.register import register

from . import BaseScheduler, BaseSchedulerConfig


@cregister("scheduler", "cosine_warmup")
class CosineWarmupScheduleConfig(BaseSchedulerConfig):
    """
    the cosine warmup scheduler
    """

    num_cycles = FloatField(
        value=0.5,
        minimum=0.0,
        suggestions=[0.5],
        help="""0.5 is for cosine annealing, the num cycles of cosine warmup, you can use the code below to check the scheduler the
            ```
                import math
                num_cycles = 0.5
                for progress in range(100):
                    progress = progress / 100
                    print(max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress))))
            ```
        """,
    )


@register("scheduler", "cosine_warmup")
class CosineWarmupSchedule(BaseScheduler):
    """CosineWarmupSchedule"""

    def __init__(
        self,
        optimizer: optim.Optimizer,
        config: CosineWarmupScheduleConfig,
        rt_config: Dict,
    ):
        super(CosineWarmupSchedule, self).__init__(optimizer, config, rt_config)
        self.config: CosineWarmupScheduleConfig

    def step_update(self, current_step: int):
        if current_step < self.config.num_warmup_steps:
            return float(current_step) / float(max(1, self.config.num_warmup_steps))
        progress = float(current_step - self.config.num_warmup_steps) / float(
            max(1, self.num_training_steps - self.config.num_warmup_steps)
        )
        return max(
            0.0,
            0.5
            * (
                1.0 + math.cos(math.pi * float(self.config.num_cycles) * 2.0 * progress)
            ),
        )

    def epoch_update(self, current_epoch: int):
        if current_epoch < self.config.num_warmup_steps:
            return float(current_epoch) / float(max(1, self.config.num_warmup_steps))
        progress = float(current_epoch - self.config.num_warmup_steps) / float(
            max(1, self.num_training_epochs - self.config.num_warmup_steps)
        )
        return max(
            0.0,
            0.5
            * (
                1.0 + math.cos(math.pi * float(self.config.num_cycles) * 2.0 * progress)
            ),
        )



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict

import torch.optim as optim
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch.optim.lr_scheduler import LambdaLR

from dlk.utils.register import register

from . import BaseScheduler, BaseSchedulerConfig


@cregister("scheduler", "linear_warmup")
class LinearWarmupScheduleConfig(BaseSchedulerConfig):
    """
    the linear warmup scheduler
    """


@register("scheduler", "linear_warmup")
class LinearWarmupSchedule(BaseScheduler):
    """linear warmup then linear decay"""

    def __init__(
        self,
        optimizer: optim.Optimizer,
        config: LinearWarmupScheduleConfig,
        rt_config: Dict,
    ):
        super(LinearWarmupSchedule, self).__init__(optimizer, config, rt_config)
        self.config: LinearWarmupScheduleConfig = config

    def step_update(self, current_step: int):
        if current_step < self.config.num_warmup_steps:
            return float(current_step) / float(max(1, self.config.num_warmup_steps))
        return max(
            0.0,
            float(self.num_training_steps - current_step)
            / float(max(1, self.num_training_steps - self.config.num_warmup_steps)),
        )

    def epoch_update(self, current_epoch: int):
        if current_epoch < self.config.num_warmup_steps:
            return float(current_epoch) / float(max(1, self.config.num_warmup_steps))
        return max(
            0.0,
            float(self.num_training_epochs - current_epoch)
            / float(max(1, self.num_training_epochs - self.config.num_warmup_steps)),
        )



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import re
from typing import Any, Callable, Dict, Tuple, Type

import torch
import torch.optim as optim
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    dataclass,
)

from dlk.utils.import_module import import_module_dir

logger = logging.getLogger(__name__)


@dataclass
class BaseOptimizerConfig(Base):
    """the base optimizer"""

    name = StrField(value="default", help="the name of default group parameters")
    optimizer_special_groups = DictField(
        value={},
        suggestions=[
            {
                "order": ["decoder", "bias"],
                "bias": {
                    "config": {"weight_decay": 0},
                    "pattern": ["bias", "LayerNorm.bias", "LayerNorm.weight"],
                },
                "decoder": {"config": {"lr": 1e-3}, "pattern": ["decoder"]},
            }
        ],
        help="""the special groups of optimizer, like
             "order": ['decoder', 'bias'], // the group order, if the para is in decoder & is in bias, set to decoder. The order name is set to the group name
             "bias": {
                 "config": { "weight_decay": 0 },
                 "pattern": ["bias",  "LayerNorm.bias", "LayerNorm.weight"]
             },
             "decoder": {
                 "config": { "lr": 1e-3 },
                 "pattern": ["decoder"]
             }""",
    )


class BaseOptimizer(object):
    def __init__(
        self,
        model: torch.nn.Module,
        config: BaseOptimizerConfig,
        optimizer: Type[optim.Optimizer],
    ):
        super(BaseOptimizer, self).__init__()
        self.config = config
        self.model = model
        self.optimizer = optimizer

    def get_optimizer(self) -> optim.Optimizer:
        """return the initialized optimizer

        Returns:
            Optimizer

        """
        return self.init_optimizer(self.optimizer, self.model, self.config)

    def init_optimizer(
        self,
        optimizer: Type[optim.Optimizer],
        model: torch.nn.Module,
        config: BaseOptimizerConfig,
    ) -> optim.Optimizer:
        """init the optimizer for paras in model, and the group is decided by config

        Args:
            optimizer: adamw, sgd, etc.
            model: pytorch model
            config: which decided the para group, lr, etc.

        Returns:
            the initialized optimizer

        """
        optimizer_special_groups = config.optimizer_special_groups
        params = []
        all_named_parameters = list(model.named_parameters())
        total_all_named_parameters = len(all_named_parameters)
        logger.info(f"All Named Params Num is {len(all_named_parameters)}")
        has_grouped_params = set()
        for special_group_name in optimizer_special_groups.get("order", []):
            group_config = optimizer_special_groups[special_group_name]["config"]
            group_patterns = optimizer_special_groups[special_group_name]["pattern"]
            # convert to regex
            combine_patterns = []
            for pattern in group_patterns:
                combine_patterns.append(f"({pattern})")
            cc_patterns = re.compile("|".join(combine_patterns))
            group_param = []
            for n, p in all_named_parameters:
                # logger.info(f"Param name {n}")
                if n in has_grouped_params:
                    continue
                if cc_patterns.search(n):  # use regex
                    has_grouped_params.add(n)
                    group_param.append(p)
            group_config["params"] = group_param
            group_config["name"] = special_group_name
            params.append(group_config)

        reserve_params = [
            p for n, p in all_named_parameters if not n in has_grouped_params
        ]
        params.append({"params": reserve_params, "name": config.name})
        logger.info(f"Param Group Nums {len(params)}")
        total_param = 0
        for group in params:
            total_param = total_param + len(group["params"])
        assert total_param == total_all_named_parameters
        opt_args = config._to_dict(only_para=True)
        opt_args.pop("name")
        opt_args.pop("optimizer_special_groups")

        return optimizer(params=params, **opt_args)

    def __call__(self):
        """the same as self.get_optimizer()"""
        return self.get_optimizer()


optimizer_dir = os.path.dirname(__file__)
import_module_dir(optimizer_dir, "dlk.optimizer")



# Copyright the author(s) of DLK.
#
# This source code is licensed under the Apache license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict

import torch
import torch.nn as nn
from intc import (
    MISSING,
    AnyField,
    Base,
    BoolField,
    DictField,
    FloatField,
    IntField,
    ListField,
    NestField,
    StrField,
    SubModule,
    cregister,
)
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from dlk.utils.register import register

from . import Module


@cregister("module", "lstm")
class LSTMConfig:
    """the lstm config"""

    bidirectional = BoolField(value=True, help="whether to use bidirectional lstm")
    input_size = IntField(value=MISSING, minimum=0, help="the input size")
    output_size = IntField(value=MISSING, minimum=0, help="the output size")
    num_layers = IntField(value=1, minimum=1, help="the number of layers")
    dropout = FloatField(value=0.0, minimum=0.0, maximum=1.0, help="the dropout rate")
    dropout_last = BoolField(
        value=True, help="whether to dropout the last layer output"
    )


@register("module", "lstm")
class LSTM(Module):
    "A wrap for nn.LSTM"

    def __init__(self, config: LSTMConfig):
        super(LSTM, self).__init__()
        self.config = config

        if self.config.num_layers <= 1:
            inlstm_dropout = 0
        else:
            inlstm_dropout = self.config.dropout

        hidden_size = self.config.output_size
        if self.config.bidirectional:
            assert self.config.output_size % 2 == 0
            hidden_size = self.config.output_size // 2

        self.lstm = nn.LSTM(
            input_size=self.config.input_size,
            hidden_size=hidden_size,
            num_layers=self.config.num_layers,
            batch_first=True,
            bidirectional=self.config.bidirectional,
            dropout=inlstm_dropout,
        )
        self.dropout_last = nn.Dropout(
            p=float(self.config.dropout) if self.config.dropout_last else 0
        )

    def forward(self, input: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns:
            lstm output the shape is the same as input

        """
        max_seq_len = input.size(1)
        seq_lens = mask.sum(1).cpu()
        pack_seq_rep = pack_padded_sequence(
            input=input, lengths=seq_lens, batch_first=True, enforce_sorted=False
        )
        pack_seq_rep = self.lstm(pack_seq_rep)[0]
        output, _ = pad_packed_sequence(
            sequence=pack_seq_rep, batch_first=True, total_length=max_seq_len
        )

        return self.dropout_last(output)
